"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import ray
import copy
import time

import pyomo.environ as pyomo

from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
from pyomo.opt import SolverStatus, TerminationCondition

from pycity_scheduling.classes import CityDistrict, Building
from pycity_scheduling.util import extract_pyomo_values
from pycity_scheduling.exceptions import NonoptimalError


def list_into_n_chunks(lst, n):
    for i in range(0, n):
        yield lst[i::n]


def get_beta(params, entity):
    """Returns the beta value for a specific entity"""
    beta = params["beta"]
    if isinstance(beta, dict):
        return beta.get(entity.id, 1.0)
    if isinstance(entity, CityDistrict):
        return 1.0
    return beta


class RaySolverNode():
    """Base class for a Ray Node which can be used to solve all entities provided to it with in a ray cluster.

    Provides an abstraction layer for algorithms, so entities can be
    assigned to nodes and optimized easily.

    Parameters
    ----------
    solver : str
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory. In addition to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.
    entities : list
        List of entities which should be optimized by this node.
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    """
    def __init__(self, solver, solver_options, entities, node_index = 0, op_horizon=0, mode="convex", robustness=None):
        self.solvers = np.empty(len(entities), dtype=object)
        for i, entity in enumerate(entities):
            self.solvers[i] = pyomo.SolverFactory(solver,  **solver_options.get("__call__", {}))
        self.solver_options = solver_options
        self.is_persistent = False
        if len(self.solvers) > 0:
            self.is_persistent = isinstance(self.solvers[0], PersistentSolver)
        self.robustness = robustness
        self.entities = []
        for entity in entities:
            self.entities.append(copy.deepcopy(entity))
        self.mode = mode
        self.model = np.empty(len(entities), dtype=object)
        self.node_index = node_index
        self.op_horizon = op_horizon

        self._reset()
        self._prepare()

    def _reset(self):
        """Reset, i.e., 'unpopulate', the previous optimization model"""
        for entity in self.entities:
            if isinstance(entity, CityDistrict):
                city_district = entity
                city_district.model = None
                for bd in city_district.get_lower_entities():
                    bd.model = None
                    if isinstance(bd, Building) and bd.has_bes:
                        bd.bes.model = None
                        for e in bd.bes.get_lower_entities():
                            e.model = None
                    if isinstance(bd, Building) and len(bd.apartments) == 1:
                        bd.apartments[0].model = None
                        for e in bd.apartments[0].get_lower_entities():
                            e.model = None
                    elif isinstance(bd, Building) and len(bd.apartments) > 1:
                        for ap in bd.apartments:
                            ap.model = None
                            for e in ap.get_lower_entities():
                                e.model = None
        return

    def _prepare(self):
        """Create the pyomo model for the entities and populate it"""
        for i, entity in enumerate(self.entities):
            model = pyomo.ConcreteModel()
            self._prepare_model(entity, model, robustness=self.robustness)
            self.model[i] = model
        return

    def _prepare_model(self, entity, model, robustness=None):
        """Add a single entity to a model"""
        if isinstance(entity, Building):
            entity.populate_model(model, self.mode, robustness=robustness)
            entity.update_model(self.mode, robustness=robustness)
        else:
            entity.populate_model(model, self.mode)
            entity.update_model(self.mode)
        return

    def set_model_betas(self, params):
        """Sets the model betas for every entity on the remote node"""
        for i, entity in enumerate(self.entities): 
            beta = get_beta(params, entity)
            self.model[i].beta = beta

    def full_update(self, robustness=None):
        """Execute the update_model function and propagate other model changes.

        Parameters
        ----------
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        """
        for i, entity in enumerate(self.entities):
            if isinstance(entity, Building):
                entity.update_model(mode=self.mode, robustness=robustness)
            else:
                entity.update_model(mode=self.mode)
            if self.is_persistent:
                self.solvers[i].set_instance(self.model[i], **self.solver_options.get("set_instance", {}))
        return

    def constr_update(self):
        """Only propagate the constraints update of the model"""
        for i, entity in enumerate(self.entities):
            if self.is_persistent:
                self.solvers[i].set_instance(self.model[i], **self.solver_options.get("set_instance", {}))
            else:
                pass
        return

    def obj_update(self, index):
        """Only propagate the objective value update of the model"""
        if self.is_persistent:
            self.solvers[index].set_objective(self.model[index].o)
        else:
            pass
        return

    def postsolve(self):
        """In this step other post-processing can be done"""
        # Updates the schedules of each asset and entity on the node
        for entity in self.entities:
            entity.update_schedule()
        return

    def get_all_schedules(self):
        """Retrives all schedules of each entity and asset on the node"""
        local_schedules = dict()
        for entity in self.entities:
            # only update the child entities when the entity is not a city district
            if isinstance(entity, CityDistrict):
                if hasattr(entity, "id") and hasattr(entity, "schedules"):
                    local_schedules.update({entity.id: entity.schedules})
            else:  
                for asset in entity.get_all_entities(): 
                    if hasattr(asset, "id") and hasattr(asset, "schedules"):
                        local_schedules.update({asset.id: asset.schedules})
        return local_schedules    

    def get_objective(self):
        obj_value = 0
        for i, entity in enumerate(self.entities):
            obj_value += pyomo.value(entity.get_objective())
        return obj_value

    def exit(self):
        ray.actor.exit_actor()
        return


@ray.remote(scheduling_strategy="SPREAD", num_cpus=1, num_gpus=0)
class RayDualDecompositionSolverNode(RaySolverNode):
    """Ray Node which can be used to solve all entities provided to it with in a ray cluster using the Dual
    Decompositon method.

    Provides an abstraction layer for algorithms, so entities can be
    assigned to nodes and optimized easily.

    Parameters
    ----------
    solver : str
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory. In addition to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.
    entities : list
        List of entities which should be optimized by this node.
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    """
    def __init__(self, solver, solver_options, entities, node_index = 0, op_horizon=0, mode="convex", robustness=None):
        super().__init__(solver, solver_options, entities, node_index, op_horizon, mode, robustness)
        self.calc_model_params()

    def calc_model_params(self):
        """ Create the model parameters locally on the actor """
        for i, entity in enumerate(self.entities): 
            beta = pyomo.Param(mutable=True, initialize=1)
            lambdas = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            self.model[i].beta = beta
            self.model[i].lambdas = lambdas

    def _set_model_lambdas(self, lambdas, index):
        """Sets the model lambdas for every entity on the remote node"""
        for t in range(self.op_horizon):
            self.model[index].lambdas[t] = lambdas[t]

    def add_objective(self, global_entity_indices):
        """Adds the objective to all models of every entity on the remote node"""
        for i, entity in enumerate(self.entities):
            global_index = global_entity_indices[entity.id]
            obj = self.model[i].beta * entity.get_objective()
            if global_index == 0:
                # penalty term is expanded and constant is omitted
                # invert sign of p_el_schedule and p_el_vars (omitted for quadratic term)
                for t in range(self.op_horizon):
                    obj -= self.model[i].lambdas[t] * entity.model.p_el_vars[t]
            else:
                for t in range(self.op_horizon):
                   obj += self.model[i].lambdas[t] * entity.model.p_el_vars[t]
            self.model[i].o = pyomo.Objective(expr=obj)
        return

    def _solve_single(self, lambdas, index, debug=True):
        """Call the solver to solve this nodes optimization problem for a single entity"""
        self._set_model_lambdas(lambdas, index)
        self.obj_update(index)
        model = self.model[index]
        entity = self.entities[index]
        solver = self.solvers[index]

        solve_options = self.solver_options.get("solve", {})
        if self.is_persistent:
            result = solver.solve(**solve_options)
        else:
            result = solver.solve(model, **solve_options)
        if (not (result.solver.termination_condition == TerminationCondition.optimal or
                result.solver.termination_condition == TerminationCondition.maxTimeLimit or
                result.solver.termination_condition == TerminationCondition.maxIterations) or
                not (result.solver.status == SolverStatus.ok or
                     result.solver.status == SolverStatus.warning or
                     result.solver.status == SolverStatus.aborted)):
            if debug:
                import pycity_scheduling.util.debug as debug
                debug.analyze_model(model, solver, result)
            raise NonoptimalError("Could not retrieve schedule from model.")

        entity.update_schedule()

        return self.entities[index].model.p_el_vars 

    def solve(self, lambdas_ref, debug=True):
        """Call the solver to solve this nodes optimization problem for every entity on the node"""
        p_el_var_list = []
        indices_list = []
        lambdas = ray.get(lambdas_ref[0])
        for i, entity in enumerate(self.entities): 
            p_el_vars = self._solve_single(lambdas, i, debug)
            p_el_var_list.append(p_el_vars)
            indices_list.append(entity.id)

        # return p_el_vars with other method or cast pyomo vars on the side of the node
        numpy_vars = np.array([extract_pyomo_values(var, float) for var in p_el_var_list])
        numpy_indices = np.array(indices_list)

        obj_value = self.get_objective()

        return [numpy_indices, numpy_vars, obj_value]

    def postsolve(self):
        """ Updates the schedules of each asset and entity on the node"""
        super().postsolve()


@ray.remote(scheduling_strategy="SPREAD", num_cpus=1, num_gpus=0)
class RayADMMSolverNode(RaySolverNode):
    """Ray Node which can be used to solve all entities provided to it with in a ray cluster using the exchange
    admm method.

    Provides an abstraction layer for algorithms, so entities can be
    assigned to nodes and optimized easily.

    Parameters
    ----------
    solver : str
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory. In addition to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.
    entities : list
        List of entities which should be optimized by this node.
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    """
    def __init__(self, solver, solver_options, entities, rho, node_index = 0, op_horizon=0, mode="convex",
                 robustness=None):
        self.rho = rho
        super().__init__(solver, solver_options, entities, node_index, op_horizon, mode, robustness)
        self.calc_model_params()
      
    def calc_model_params(self):
        """ Create the model parameters localy on the actor """
        for i, entity in enumerate(self.entities): 
            beta = pyomo.Param(mutable=True, initialize=1)
            xs_ = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            us = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            last_p_el_schedules = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            self.model[i].beta = beta
            self.model[i].xs_ = xs_
            self.model[i].us = us
            self.model[i].last_p_el_schedules = last_p_el_schedules

    def _set_model_variables(self, last_p_els, last_x_, last_u, local_index, global_index):
        """Sets the model vars for entity on the remote node"""
        for t in range(self.op_horizon):
            self.model[local_index].last_p_el_schedules[t] = last_p_els[global_index][t]
            self.model[local_index].xs_[t] = last_x_[t]
            self.model[local_index].us[t] = last_u[t]

    def add_objective(self, global_entity_indices):
        """Adds the objective to all models of every entity on the remote node"""
        for i, entity in enumerate(self.entities):
            global_index = global_entity_indices[entity.id]
            obj = self.model[i].beta * entity.get_objective()
            for t in range(self.op_horizon):
                obj += self.rho / 2 * entity.model.p_el_vars[t] * entity.model.p_el_vars[t]
            # penalty term is expanded and constant is omitted
            if global_index == 0:
                # invert sign of p_el_schedule and p_el_vars (omitted for quadratic term)
                penalty = [(-self.model[i].last_p_el_schedules[t] - self.model[i].xs_[t] - self.model[i].us[t])
                           for t in range(self.op_horizon)]
                for t in range(self.op_horizon):
                    obj += self.rho * penalty[t] * entity.model.p_el_vars[t]
            else:
                penalty = [(-self.model[i].last_p_el_schedules[t] + self.model[i].xs_[t] + self.model[i].us[t])
                           for t in range(self.op_horizon)]
                for t in range(self.op_horizon):
                    obj += self.rho * penalty[t] * entity.model.p_el_vars[t]
            self.model[i].o = pyomo.Objective(expr=obj)
        return

    def _solve_single(self, last_p_els, last_x_, last_u, local_index, global_index, debug=True):
        """Call the solver to solve this nodes optimization problem for a single entity"""
        self._set_model_variables(last_p_els, last_x_, last_u, local_index, global_index)
        self.obj_update(local_index)

        model = self.model[local_index]
        entity = self.entities[local_index]
        solver = self.solvers[local_index]

        solve_options = self.solver_options.get("solve", {})
        if self.is_persistent:
            result = solver.solve(**solve_options)
        else:
            result = solver.solve(model, **solve_options)
        if (not (result.solver.termination_condition == TerminationCondition.optimal or
                result.solver.termination_condition == TerminationCondition.maxTimeLimit or
                result.solver.termination_condition == TerminationCondition.maxIterations) or
                not (result.solver.status == SolverStatus.ok or
                     result.solver.status == SolverStatus.warning or
                     result.solver.status == SolverStatus.aborted)):
            if debug:
                import pycity_scheduling.util.debug as debug
                debug.analyze_model(model, solver, result)
            raise NonoptimalError("Could not retrieve schedule from model.")

        entity.update_schedule()

        return self.entities[local_index].model.p_el_vars 

    def solve(self, variable_refs, debug=True):
        """Call the solver to solve this nodes optimization problem for every entity on the node

        Parameters
        ----------
        variables : list of lists of variables, optional
            Can contain a list for each node in nodes to indicate to pyomo which
            variables should be loaded back into the model. Specifying this can
            lead to a significant speedup.
        debug : bool, optional
            Specify whether detailed debug information shall be printed. Defaults
            to true.
        """
        p_el_var_list = []
        indices_list = []

        last_p_els = ray.get(variable_refs[0])
        last_x_ = ray.get(variable_refs[1])
        last_u_ = ray.get(variable_refs[2])
        global_entity_indices = ray.get(variable_refs[3])

        for i, entity in enumerate(self.entities): 
            global_index = global_entity_indices[entity.id]
            p_el_vars = self._solve_single(last_p_els, last_x_, last_u_, i, global_index, debug)
            p_el_var_list.append(p_el_vars)
            indices_list.append(entity.id)

        # return p_el_vars with other method or cast pyomo vars on the side of the node
        numpy_vars = np.array([extract_pyomo_values(var, float) for var in p_el_var_list])
        numpy_indices = np.array(indices_list)

        obj_value = self.get_objective()

        return [numpy_indices, numpy_vars, obj_value]

    def postsolve(self):
        """ Updates the schedules of each asset and entity on the node """
        super().postsolve()


@ray.remote(scheduling_strategy="SPREAD", num_cpus=1, num_gpus=0)
class RayADMMMIQPSolverNode(RaySolverNode):
    """Ray Node which can be used to solve all entities provided to it with in a ray cluster using the Exchange
    MIQP ADMM method.

    Provides an abstraction layer for algorithms, so entities can be
    assigned to nodes and optimized easily.

    Parameters
    ----------
    solver : str
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory. In addition to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.
    entities : list
        List of entities which should be optimized by this node.
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    """
    def __init__(self, solver, solver_options, entities, gamma, rho, node_index=0, max_iterations=1000, op_horizon=0,
                 mode="convex", x_update_mode="constrained", robustness=None):
        self.x_update_mode = x_update_mode
        self.gamma = gamma
        self.rho = rho
        self.max_iterations = max_iterations
        super().__init__(solver, solver_options, entities, node_index, op_horizon, mode, robustness)
        self.calc_model_params()

        self.feasible = False
        self.node_binaries, self.node_bin_values, self.node_x_k_values, \
             self.node_u_k_values = self._get_binaries()

        self.node_equalities_t, self.node_equalities_n, \
             self.node_inequalities_t, self.node_inequalities_n = self._get_constraints()

    def calc_model_params(self):
        """ Create the model parameters locally on the actor """
        for i, entity in enumerate(self.entities): 
            beta = pyomo.Param(mutable=True, initialize=1)
            xs_ = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            us = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            last_p_el_schedules = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            gamma_ = pyomo.Param(mutable=True, initialize=self.gamma)
            rho_ = pyomo.Param(mutable=True, initialize=self.rho)
            self.model[i].gamma_ = gamma_
            self.model[i].rho_ = rho_
            self.model[i].beta = beta
            self.model[i].xs_ = xs_
            self.model[i].us = us
            self.model[i].last_p_el_schedules = last_p_el_schedules        

    def _get_binaries(self):
        local_binaries = []
        local_bin_values = []
        local_x_k_values = []
        local_u_k_values = []
        for i, entity in enumerate(self.entities):
            node_binaries_list = []
            node_bin_values_list = []
            node_x_k_values_list = []
            node_u_k_values_list = []
            if not isinstance(entity, CityDistrict):
                for en in entity.get_all_entities():
                    for variable in en.model.component_objects(pyomo.Var):
                        if variable[0].domain is pyomo.Binary:
                            bin_index_list = []
                            x_k_index_list = []
                            u_k_index_list = []
                            # change domain
                            variable.domain = pyomo.Reals
                            variable.setlb(0)
                            variable.setub(1)
                            node_binaries_list.append(variable)
                            for index in variable:
                                variable[index].fixed = False
                            for t in range(self.op_horizon):
                                x_value_array = np.zeros(self.max_iterations+1)
                                x_k_value_array = np.zeros(self.max_iterations+1)
                                u_k_value_array = np.zeros(self.max_iterations+1)
                                bin_index_list.append(x_value_array)
                                x_k_index_list.append(x_k_value_array)
                                u_k_index_list.append(u_k_value_array)
                            node_bin_values_list.append(bin_index_list)
                            node_x_k_values_list.append(x_k_index_list)
                            node_u_k_values_list.append(u_k_index_list)
            local_binaries.append(node_binaries_list)
            local_bin_values.append(node_bin_values_list)
            local_x_k_values.append(node_x_k_values_list)
            local_u_k_values.append(node_u_k_values_list)

        if len(local_binaries) <= 1:
            local_binaries.append([])

        if len(local_bin_values) <= 1:
            local_bin_values.append([])

        if len(local_x_k_values) <= 1:
            local_x_k_values.append([])

        if len(local_u_k_values) <= 1:
            local_u_k_values.append([])

        return np.array(local_binaries, dtype=object), np.array(local_bin_values, dtype=object), \
            np.array(local_x_k_values, dtype=object), np.array(local_u_k_values, dtype=object),

    def _get_constraints(self):
        local_equalities_t = []
        local_equalities_n = []
        local_inequalities_t = []
        local_inequalities_n = []

        for i, entity in enumerate(self.entities):
            equalities_list_t = []
            inequalities_list_t = []
            equalities_list_n = []
            inequalities_list_n = []
            equality = False
            inequality = False
            none_index = False
            if not isinstance(entity, CityDistrict):
                for en in entity.get_all_entities():
                    for constraint in en.model.component_objects(pyomo.Constraint):
                        if self.x_update_mode == "unconstrained":
                            constraint.deactivate()
                        for index in constraint:
                            if index is None:
                                none_index = True
                            # check if the constraint is an equality constraint and write it in the form Ax-b=0
                            # for each index
                            if pyomo.value(constraint[index].lower) == pyomo.value(constraint[index].upper):
                                expr = constraint[index].body - constraint[index].lower
                                constraint[index].set_value(expr == 0)
                                equality = True

                            # if the constraint is not an equality constraint it has to be an inequality constraint;
                            # the next three checks are about to write that constraint in the form Cx-d >= 0
                            elif pyomo.value(constraint[index].upper) is None:
                                expr = constraint[index].body - constraint[index].lower
                                constraint[index].set_value(expr >= 0)
                                inequality = True

                            elif pyomo.value(constraint[index].lower) is None:
                                expr = -constraint[index].body + constraint[index].upper
                                constraint[index].set_value(expr >= 0)
                                inequality = True
                            else:
                                expr = -constraint[index].body + constraint[index].upper
                                expr = constraint[index].body - constraint[index].lower
                                print("WARNING: This constraint has a lb and ub and is hence not deactivated!")
                                inequality = True
                        if equality:
                            if none_index:
                                equalities_list_n.append(constraint)
                                none_index = False
                            else:
                                equalities_list_t.append(constraint)
                            equality = False
                        if inequality:
                            if none_index:
                                inequalities_list_n.append(constraint)
                                none_index = False
                            else:
                                inequalities_list_t.append(constraint)
                            inequality = False
            local_equalities_t.append(equalities_list_t)
            local_equalities_n.append(equalities_list_n)
            local_inequalities_t.append(inequalities_list_t)
            local_inequalities_n.append(inequalities_list_n)

        if len(local_equalities_t) <= 1:
            local_equalities_t.append([])
            
        if len(local_equalities_n) <= 1:
            local_equalities_n.append([])

        if len(local_inequalities_t) <= 1:
            local_inequalities_t.append([])

        if len(local_inequalities_n) <= 1:
            local_inequalities_n.append([])

        return np.array(local_equalities_t, dtype=object), np.array(local_equalities_n, dtype=object),\
            np.array(local_inequalities_t, dtype=object), np.array(local_inequalities_n, dtype=object)

    def set_model_parameters(self):
        """Sets the model parameters for an entity on the remote node"""
        for i, entity in enumerate(self.entities):
            # Create parameters for each binary variable
            length = len(self.node_binaries[i])
            if length != 0:
                self.model[i].bin_set = pyomo.RangeSet(0, length - 1)
                self.model[i].x_k = pyomo.Param(self.model[i].bin_set, entity.model.t, mutable=True,
                                                initialize=0)
                self.model[i].u_xk = pyomo.Param(self.model[i].bin_set, entity.model.t, mutable=True,
                                                 initialize=0)
            if self.x_update_mode == "unconstrained":
                # Create parameters for each constraint
                length = len(self.node_equalities_t[i])
                if length != 0:
                    self.model[i].eq_t_set = pyomo.RangeSet(0, length - 1)
                    self.model[i].u_eq_t = pyomo.Param(self.model[i].eq_t_set, entity.model.t, mutable=True,
                                                       initialize=0)

                length = len(self.node_equalities_n[i])
                if length != 0:
                    self.model[i].eq_n_set = pyomo.RangeSet(0, length-1)
                    self.model[i].u_eq_n = pyomo.Param(self.model[i].eq_n_set, mutable=True, initialize=0)

                length = len(self.node_inequalities_t[i])
                if length != 0:
                    self.model[i].ineq_t_set = pyomo.RangeSet(0, length - 1)
                    self.model[i].u_ineq_t = pyomo.Param(self.model[i].ineq_t_set, entity.model.t,
                                                         mutable=True, initialize=0)
                    self.model[i].v_k_t = pyomo.Param(self.model[i].ineq_t_set, entity.model.t, mutable=True,
                                                      initialize=0)

                length = len(self.node_inequalities_n[i])
                if length != 0:
                    self.model[i].ineq_n_set = pyomo.RangeSet(0, length - 1)
                    self.model[i].u_ineq_n = pyomo.Param(self.model[i].ineq_n_set, mutable=True, initialize=0)
                    self.model[i].v_k_n = pyomo.Param(self.model[i].ineq_n_set, mutable=True, initialize=0)
        return

    def _set_model_variables(self, last_p_els, last_x_, last_u, local_index, global_index):
        """Sets the model vars for an entity on the remote node"""
        for t in range(self.op_horizon):
            self.model[local_index].last_p_el_schedules[t] = last_p_els[global_index][t]
            self.model[local_index].xs_[t] = last_x_[t]
            self.model[local_index].us[t] = last_u[t]

    def add_objective(self, exchange_admm_obj_terms=True, miqp_admm_obj_terms=True):
        """Adds the objective function for an entity on the remote node"""
        for i, entity in enumerate(self.entities):
            obj = self.model[i].beta * entity.get_objective()
            for t in range(entity.op_horizon):
                obj += self.model[i].rho_ / 2 * entity.model.p_el_vars[t] * entity.model.p_el_vars[t]

            # In the following, add the additional expressions to solve the sub-problems by Exchange ADMM.
            if exchange_admm_obj_terms:
                # penalty term is expanded and constant is omitted
                if isinstance(entity, CityDistrict):# i == 0:
                    # invert sign of p_el_schedule and p_el_vars (omitted for quadratic term)
                    penalty = [(-self.model[i].last_p_el_schedules[t] - self.model[i].xs_[t] - self.model[i].us[t])
                                for t in range(entity.op_horizon)]
                    for t in range(entity.op_horizon):
                        obj += self.model[i].rho_ * penalty[t] * entity.model.p_el_vars[t]
                else:
                    penalty = [(-self.model[i].last_p_el_schedules[t] + self.model[i].xs_[t] + self.model[i].us[t])
                                for t in range(entity.op_horizon)]
                    for t in range(entity.op_horizon):
                        obj += self.model[i].rho_ * penalty[t] * entity.model.p_el_vars[t]
            # In the following, add the additional expressions to solve the sub-problems by MIQP ADMM.
            if miqp_admm_obj_terms:
                # binary variables contribution
                length = len(self.node_binaries[i])
                if length != 0:
                    for x, k in zip(self.node_binaries[i], range(length)):
                        obj += self.model[i].rho_ / 2 * sum((x[t] - self.model[i].x_k[k, t] + self.model[i].gamma_ *
                                                             self.model[i].u_xk[k, t])**2
                                                            for t in range(self.op_horizon))

                if self.x_update_mode == "unconstrained":
                    # add the contributions of the constraints (time and none indexed)
                    length = len(self.node_equalities_t[i])
                    if length != 0:
                        for eq_constr, k in zip(self.node_equalities_t[i], range(length)):
                            obj += self.model[i].rho_ / 2 * sum((eq_constr[t].body + self.model[i].u_eq_t[k, t]) ** 2
                                                                for t in range(len(eq_constr)))

                    length = len(self.node_equalities_n[i])
                    if length != 0:
                        for eq_constr, k in zip(self.node_equalities_n[i], range(length)):
                            obj += self.model[i].rho_ / 2 * (eq_constr.body + self.model[i].u_eq_n[k]) ** 2

                    length = len(self.node_inequalities_t[i])
                    if length != 0:
                        for ineq_constr, k in zip(self.node_inequalities_t[i], range(length)):
                            obj += self.model[i].rho_ / 2 * sum((ineq_constr[t].body + self.model[i].u_ineq_t[k, t] -
                                                                 self.model[i].v_k_t[k, t]) ** 2
                                                                for t in range(len(ineq_constr)))

                    length = len(self.node_inequalities_n[i])
                    if length != 0:
                        for ineq_constr, k in zip(self.node_inequalities_n[i], range(length)):
                            obj += self.model[i].rho_ / 2 * (ineq_constr.body + self.model[i].u_ineq_n[k] -
                                                             self.model[i].v_k_n[k])**2

            # if we want to redefine the objective for a certain node, then we should first reset the old objective
            try:
                self.model[i].del_component(self.model[i].o)
            except AttributeError:
                pass
            self.model[i].o = pyomo.Objective(expr=obj)
        return

    def _solve_single(self, last_p_els, last_x_, last_u, local_index, global_index, debug=True):
        """Call the solver to solve this nodes optimization problem for a single entity"""
        self._set_model_variables(last_p_els, last_x_, last_u, local_index, global_index)
        self.obj_update(local_index)

        model = self.model[local_index]
        entity = self.entities[local_index]
        solver = self.solvers[local_index]

        solve_options = self.solver_options.get("solve", {})
        if self.is_persistent:
            result = solver.solve(**solve_options)
        else:
            result = solver.solve(model, **solve_options)
        if (not (result.solver.termination_condition == TerminationCondition.optimal or
                result.solver.termination_condition == TerminationCondition.maxTimeLimit or
                result.solver.termination_condition == TerminationCondition.maxIterations) or
                not (result.solver.status == SolverStatus.ok or
                     result.solver.status == SolverStatus.warning or
                     result.solver.status == SolverStatus.aborted)):
            if debug:
                import pycity_scheduling.util.debug as debug
                debug.analyze_model(model, solver, result)
            raise NonoptimalError("Could not retrieve schedule from model.")

        entity.update_schedule()

        return self.entities[local_index].model.p_el_vars 

    def solve(self, variable_refs, update_constr=False, debug=True):
        """Call the solver to solve this nodes optimization problem for every entity on the node"""
        p_el_var_list = []
        indices_list = []

        last_p_els = ray.get(variable_refs[0])
        last_x_ = ray.get(variable_refs[1])
        last_u_ = ray.get(variable_refs[2])
        global_entity_indices = ray.get(variable_refs[3])

        if update_constr:
            self.constr_update()

        is_feasible = True
        for i, entity in enumerate(self.entities): 
            global_index = global_entity_indices[entity.id]
            try:
                p_el_vars = self._solve_single(last_p_els, last_x_, last_u_, i, global_index, debug)
                p_el_var_list.append(p_el_vars)
                indices_list.append(entity.id)
            except:
                is_feasible = False

        # return p_el_vars with other method or cast pyomo vars on the side of the node
        numpy_vars = np.array([extract_pyomo_values(var, float) for var in p_el_var_list])
        numpy_indices = np.array(indices_list)

        obj_value = self.get_objective()
        return [numpy_indices, numpy_vars, obj_value, is_feasible]

    def incentive_signal_update(self, iteration_num):
        """Update all MIQP ADMM parameters (first the constraints then the variables) for every entity on the node"""
        for i, entity in enumerate(self.entities):
            if self.x_update_mode == "unconstrained":
                length = len(self.node_inequalities_t[i])
                if length != 0:
                    for ineq_constr, k in zip(self.node_inequalities_t[i], range(length)):
                        for t in range(len(ineq_constr)):
                            v_update = pyomo.value(ineq_constr[t].body) + self.model[i].u_ineq_t[k, t].value
                            if v_update > 0:
                                self.model[i].v_k_t[k, t] = v_update
                            else:
                                self.model[i].v_k_t[k, t] = 0

                length = len(self.node_inequalities_n[i])
                if length != 0:
                    for ineq_constr, k in zip(self.node_inequalities_n[i], range(length)):
                        v_update = pyomo.value(ineq_constr.body) + self.model[i].u_ineq_n[k].value
                        if v_update > 0:
                            self.model[i].v_k_n[k] = v_update
                        else:
                            self.model[i].v_k_n[k] = 0

                length = len(self.node_inequalities_t[i])
                if length != 0:
                    for ineq_constr, k in zip(self.node_inequalities_t[i], range(length)):
                        for t in range(len(ineq_constr)):
                            self.model[i].u_ineq_t[k, t] = self.model[i].u_ineq_t[k, t].value + \
                                                        pyomo.value(ineq_constr[t].body) - \
                                                        self.model[i].v_k_t[k, t].value

                length = len(self.node_inequalities_n[i])
                if length != 0:
                    for ineq_constr, k in zip(self.node_inequalities_n[i], range(length)):
                        self.model[i].u_ineq_n[k] = self.model[i].u_ineq_n[k].value + pyomo.value(ineq_constr.body) - \
                                                    self.model[i].v_k_n[k].value

                length = len(self.node_equalities_t[i])
                if length != 0:
                    for eq_constr, k in zip(self.node_equalities_t[i], range(length)):
                        for t in range(len(eq_constr)):
                            self.model[i].u_eq_t[k, t] = (self.model[i].u_eq_t[k, t].value +
                                                          pyomo.value(eq_constr[t].body))

                length = len(self.node_equalities_n[i])
                if length != 0:
                    for eq_constr, k in zip(self.node_equalities_n[i], range(length)):
                        self.model[i].u_eq_n[k] = self.model[i].u_eq_n[k].value + pyomo.value(eq_constr.body)

            # x_k update
            length = len(self.node_binaries[i])
            if length != 0:
                for x, k in zip(self.node_binaries[i], range(length)):
                    for t in range(self.op_horizon):
                        # Integer rounding:
                        self.model[i].x_k[k, t] = round(x[t].value + self.model[i].u_xk[k, t].value)
                        # Binary rounding:
                        # x_update = abs(x[t].value + node.model.u_xk[k, t].value)
                        # if x_update >= 0.5:
                        #    node.model.x_k[k, t] = 1
                        # else:
                        #    node.model.x_k[k, t] = 0
                        self.node_x_k_values[i][k][t][iteration_num] = self.model[i].x_k[k, t].value

            # u_xk update
            length = len(self.node_binaries[i])
            if length != 0:
                for x, x_val, u_k, k in zip(self.node_binaries[i],self.node_bin_values[i],
                                            self.node_u_k_values[i], range(length)):
                    for t in range(self.op_horizon):
                       
                        self.model[i].u_xk[k, t] = self.model[i].u_xk[k, t].value + x[t].value - \
                                                self.model[i].x_k[k, t].value
                        x_val[t][iteration_num] = x[t].value
                        u_k[t][iteration_num] = self.model[i].u_xk[k, t].value
        # Return local x_ks and u_ks to head to update district_x_k_values[i] and u_k[t] on head side
        return

    def fix_variables(self):
        """Fix the binary variables for all entities on the remote node"""
        for i, entity in enumerate(self.entities):
            length = len(self.node_binaries[i])
            if length != 0:
                for x, k in zip(self.node_binaries[i], range(length)):
                    for t in range(self.op_horizon):
                        x_k = self.model[i].x_k[k, t].value
                        x[t].fix(x_k)
        return

    def release_variables(self):
        """Release the binary variables for all entities on the remote node"""
        for i, entity in enumerate(self.entities):
            length = len(self.node_binaries[i])
            if length != 0:
                for x, k in zip(self.node_binaries[i], range(length)):
                    for t in range(self.op_horizon):
                        x[t].unfix()
        return

    def postsolve(self):
        """Updates the schedules of each asset and entity on the node"""
        super().postsolve()
        pass

    def update_gamma(self, new_gamma):
        """Updates the gamma parameter for all entities on the remote node"""
        self.gamma = new_gamma
        for i, entity in enumerate(self.entities):
            self.model[i].gamma_ = new_gamma
        return

    def update_rho(self, new_rho):
        """Updates the rho parameter for all entities on the remote node"""
        self.rho = new_rho
        for i, entity in enumerate(self.entities):
            self.model[i].rho_ = new_rho
        return

    def activate_constraints(self):
        """Activates all constraints for all entities on the remote node"""
        for i, entity in enumerate(self.entities):
            if not isinstance(entity, CityDistrict):
                for en in entity.get_all_entities():
                    for constraint in en.model.component_objects(pyomo.Constraint):
                        constraint.activate()
        return

    def deactivate_constraints(self):
        """Deactivates all constraints for all entities on the remote node"""
        for i, entity in enumerate(self.entities):
            if not isinstance(entity, CityDistrict):
                for en in entity.get_all_entities():
                    for constraint in en.model.component_objects(pyomo.Constraint):
                        constraint.deactivate()
        return

    # Returns True, if no constraint or binary variable violations occur
    def check_violations(self, detailed_print_out=False):
        """Checks constraint and binary variable violations for all entities on the remote node"""
        constr_counter = 0
        var_counter = 0
        number_constr = 0
        number_var = 0
        for i, entity in enumerate(self.entities):
            length = len(self.node_binaries[i])
            if length != 0:
                for x, val, x_k, u_k in zip(self.node_binaries[i], self.node_bin_values[i],
                                            self.node_x_k_values[i], self.node_u_k_values[i]):
                    for t in range(self.op_horizon):
                        if abs(x[t].value) < 0.1 or (0.9 < x[t].value < 1.1):
                            pass
                        else:
                            var_counter += 1
                            if detailed_print_out:
                                x[t].pprint()
                        number_var += 1

            length = len(self.node_equalities_t[i])
            if length != 0:
                for eq_constr in self.node_equalities_t[i]:
                    for t in range(len(eq_constr)):
                        if abs(pyomo.value(eq_constr[t].body)) > 0.05:
                            if detailed_print_out:
                                print(eq_constr[t].body, " ", pyomo.value(eq_constr[t].body))
                            constr_counter += 1
                        number_constr += 1

            length = len(self.node_equalities_n[i])
            if length != 0:
                for eq_constr in self.node_equalities_n[i]:
                    if abs(pyomo.value(eq_constr.body)) > 0.05:
                        if detailed_print_out:
                            print(eq_constr.body, " ", pyomo.value(eq_constr.body))
                        constr_counter += 1
                    number_constr += 1

            length = len(self.node_inequalities_t[i])
            if length != 0:
                for ineq_constr in self.node_inequalities_t[i]:
                    for t in range(len(ineq_constr)):
                        if pyomo.value(ineq_constr[t].body) < -0.05:
                            if detailed_print_out:
                                print(ineq_constr[t].body, " ", pyomo.value(ineq_constr[t].body))
                            constr_counter += 1
                        number_constr += 1

            length = len(self.node_inequalities_n[i])
            if length != 0:
                for ineq_constr in self.node_inequalities_n[i]:
                    if pyomo.value(ineq_constr.body) < -0.05:
                        if detailed_print_out:
                            print(ineq_constr.body, " ", pyomo.value(ineq_constr.body))
                        constr_counter += 1
                    number_constr += 1
        return constr_counter, var_counter, number_constr, number_var


@ray.remote(scheduling_strategy="SPREAD", num_cpus=1, num_gpus=0)
class RayDerivativeFreeALADINSolverNode(RaySolverNode):
    """Ray Node which can be used to solve all entities provided to it with in a ray cluster using the derivative-free
     ALADIN method.

    Provides an abstraction layer for algorithms, so entities can be
    assigned to nodes and optimized easily.

    Parameters
    ----------
    solver : str
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory. In addition to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.
    entities : list
        List of entities which should be optimized by this node.
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    """

    def __init__(self, solver, solver_options, entities, rho, hessian_scaling, node_index=0, op_horizon=0, mode="convex",
                 robustness=None):
        self.rho = rho
        self.hessian_scaling = hessian_scaling
        super().__init__(solver, solver_options, entities, node_index, op_horizon, mode, robustness)
        self.calc_model_params()

    def calc_model_params(self):
        """ Create the model parameters localy on the actor """
        for i, entity in enumerate(self.entities):
            beta = pyomo.Param(mutable=True, initialize=1)
            rho = pyomo.Param(mutable=True, initialize=self.rho)
            lambda_k = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            x_k = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            self.model[i].beta = beta
            self.model[i].rho = rho
            self.model[i].lambda_k = lambda_k
            self.model[i].x_k = x_k

    def _set_model_variables(self, lambda_k, x_k, local_index, global_index):
        """Sets the model vars for entity on the remote node"""
        for t in range(self.op_horizon):
            self.model[local_index].lambda_k[t] = lambda_k[t]
            self.model[local_index].x_k[t] = x_k[global_index, t]

    def add_objective(self):
        """Adds the objective to all models of every entity on the remote node"""
        for i, entity in enumerate(self.entities):
            obj = self.model[i].beta * entity.get_objective()
            # ToDo: Attention - Currently only defined for "price", "co2", "peak-shaving" and "least-squares"
            decision_vars = entity.get_decision_var()

            # lambda penalty term
            for t in range(self.op_horizon):
                obj += self.model[i].lambda_k[t] * decision_vars[t]

            # augmented penalty term
            # ToDo: Define custom scaling...
            scaling_i = self.hessian_scaling * np.eye(self.op_horizon)
            penalty = [decision_vars[t] - self.model[i].x_k[t] for t in range(self.op_horizon)]
            for j in range(self.op_horizon):
                for k in range(self.op_horizon):
                    if scaling_i[j][k] != 0.0:
                        obj += self.model[i].rho * (penalty[j] * scaling_i[j][k] * penalty[k])
            self.model[i].o = pyomo.Objective(expr=obj)
        return

    def _solve_single(self, lambda_k, x_k, local_index, global_index, debug=True):
        """Call the solver to solve this nodes optimization problem for a single entity

        Parameters
        ----------
        variables : list of lists of variables, optional
            Can contain a list for each node in nodes to indicate to pyomo which
            variables should be loaded back into the model. Specifying this can
            lead to a significant speedup.
        debug : bool, optional
            Specify whether detailed debug information shall be printed. Defaults
            to true.
        """
        self._set_model_variables(lambda_k, x_k, local_index, global_index)
        self.obj_update(local_index)

        model = self.model[local_index]
        entity = self.entities[local_index]
        solver = self.solvers[local_index]

        solve_options = self.solver_options.get("solve", {})
        if self.is_persistent:
            result = solver.solve(**solve_options)
        else:
            result = solver.solve(model, **solve_options)
        if result.solver.termination_condition != TerminationCondition.optimal or \
                result.solver.status != SolverStatus.ok:
            if debug:
                import pycity_scheduling.util.debug as debug
                debug.analyze_model(model, solver, result)
            raise NonoptimalError("Could not retrieve schedule from model.")

        entity.update_schedule()

        return self.entities[local_index].model.p_el_vars

    def solve(self, variable_refs, debug=True):
        """Call the solver to solve this nodes optimization problem for every entity on the node

        Parameters
        ----------
        variables : list of lists of variables, optional
            Can contain a list for each node in nodes to indicate to pyomo which
            variables should be loaded back into the model. Specifying this can
            lead to a significant speedup.
        debug : bool, optional
            Specify whether detailed debug information shall be printed. Defaults
            to true.
        """
        p_el_var_list = []
        indices_list = []
        sub_times_list = []

        lambda_k = ray.get(variable_refs[0])
        x_k = ray.get(variable_refs[1])
        global_entity_indices = ray.get(variable_refs[2])

        for i, entity in enumerate(self.entities):
            global_index = global_entity_indices[entity.id]
            t_0 = time.monotonic()
            p_el_vars = self._solve_single(lambda_k, x_k, i, global_index, debug)
            t_1 = time.monotonic()
            sub_times_list.append(t_1 - t_0)
            p_el_var_list.append(p_el_vars)
            indices_list.append(entity.id)

        # return p_el_vars with other method or cast pyomo vars on the side of the node
        numpy_vars = np.array([extract_pyomo_values(var, float) for var in p_el_var_list])
        numpy_indices = np.array(indices_list)
        numpy_times = np.array(sub_times_list)

        obj_value = self.get_objective()

        return [numpy_indices, numpy_vars, obj_value, numpy_times]

    def postsolve(self):
        """ Updates the schedules of each asset and entity on the node """
        super().postsolve()



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import unittest
import pyomo.environ as pyomo

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import algorithms
from pycity_scheduling.exceptions import *


class TestAlgorithms(unittest.TestCase):
    def setUp(self):
        t = Timer(op_horizon=2)
        p = Prices(t)
        w = Weather(t)
        e = Environment(t, w, p)
        cd = CityDistrict(e, objective='peak-shaving')

        bd1 = Building(e, objective='peak-shaving')
        cd.addEntity(bd1, [0, 0])
        bes = BuildingEnergySystem(e)
        bd1.addEntity(bes)
        ths = ThermalHeatingStorage(e, 40, 0.5)
        bes.addDevice(ths)
        eh = ElectricHeater(e, 10)
        bes.addDevice(eh)
        ap = Apartment(e)
        bd1.addEntity(ap)
        load = np.array([10, 10])
        fi = FixedLoad(e, method=0, demand=load)
        ap.addEntity(fi)
        sh = SpaceHeating(e, method=0, loadcurve=load)
        ap.addEntity(sh)

        bd2 = Building(e, objective='peak-shaving')
        cd.addEntity(bd2, [0, 0])
        bes = BuildingEnergySystem(e)
        bd2.addEntity(bes)
        ths = ThermalHeatingStorage(e, 40, 0.5)
        bes.addDevice(ths)
        eh = ElectricHeater(e, 20)
        bes.addDevice(eh)
        ap = Apartment(e)
        bd2.addEntity(ap)
        load = np.array([20, 20])
        fi = FixedLoad(e, method=0, demand=load)
        ap.addEntity(fi)
        sh = SpaceHeating(e, method=0, loadcurve=load)
        ap.addEntity(sh)

        wec = WindEnergyConverter(e, np.array([50, 50]), np.array([0, 0]), force_renewables=False)
        cd.addEntity(wec, [0, 0])

        pv = Photovoltaic(e, method=1, peak_power=4.6, force_renewables=False)
        cd.addEntity(pv, [0, 0])

        self.timer = t
        self.cd = cd
        self.bd1 = bd1
        self.bd2 = bd2
        return

    def test_exchange_admm(self):
        f = algorithms['exchange-admm'](self.cd, rho=2.0, eps_primal=0.001)
        r = f.solve()

        self.assertAlmostEqual(20, self.bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(20, self.bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[0], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[1], 4)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[1], 2)
        self.assertTrue(r["r_norms"][-2] > 0.001 or r["s_norms"][-2] > 1)
        self.assertGreater(0.001, r["r_norms"][-1])
        self.assertGreater(1, r["s_norms"][-1])

        # Test infeasible model:
        self.bd1.model.new_constr = pyomo.Constraint(expr=self.bd1.model.p_el_vars[0] == 0)
        self.bd1.model.p_el_vars[0].setub(15.0)
        with self.assertRaises(NonoptimalError):
            f.solve(full_update=True, debug=False)
        return

    def test_exchange_admm_beta(self):
        t = Timer(op_horizon=2)
        p = Prices(t)
        w = Weather(t)
        e = Environment(t, w, p)
        cd = CityDistrict(e, objective='peak-shaving')

        bd1 = Building(e, objective='none')
        cd.addEntity(bd1, [0, 0])
        bes = BuildingEnergySystem(e)
        bd1.addEntity(bes)
        ap = Apartment(e)
        bd1.addEntity(ap)
        load = np.array([10, -10])
        fi = FixedLoad(e, method=0, demand=load)
        ap.addEntity(fi)

        bd2 = Building(e, objective='peak-shaving')
        cd.addEntity(bd2, [0, 0])
        bes = BuildingEnergySystem(e)
        bd2.addEntity(bes)
        bat = Battery(e, 100, 100, eta=1.0, storage_end_equality=True)
        bes.addDevice(bat)

        f = algorithms['exchange-admm'](cd, rho=2, eps_primal=0.001, eps_dual=0.01)
        r = f.solve()
        self.assertAlmostEqual(10, bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(-10, bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(-5, bd2.p_el_schedule[0], 2)
        self.assertAlmostEqual(5, bd2.p_el_schedule[1], 2)
        self.assertAlmostEqual(5, cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(-5, cd.p_el_schedule[1], 2)

        r = f.solve(full_update=False, beta=0)
        self.assertAlmostEqual(10, bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(-10, bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(-10, bd2.p_el_schedule[0], 2)
        self.assertAlmostEqual(10, bd2.p_el_schedule[1], 2)
        self.assertAlmostEqual(0, cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(0, cd.p_el_schedule[1], 2)

        r = f.solve(full_update=False, beta={bd1.id: 0, bd2.id: 1, cd.id: 0})
        self.assertAlmostEqual(10, bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(-10, bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(0, bd2.p_el_schedule[0], 2)
        self.assertAlmostEqual(0, bd2.p_el_schedule[1], 2)
        self.assertAlmostEqual(10, cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(-10, cd.p_el_schedule[1], 2)
        return

    def test_exchange_miqp_admm(self):
        # Exchange MIQP ADMM constrained:
        f = algorithms['exchange-miqp-admm'](self.cd, rho=2.0, eps_exch_primal=0.001, eps_exch_dual=0.01,
                                             x_update_mode='constrained', mode="integer")
        r = f.solve()

        self.assertAlmostEqual(20, self.bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(20, self.bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[0], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[1], 4)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[1], 2)

        self.assertGreater(0.001, r["r_norms"][-1])
        self.assertGreater(0.01, r["s_norms"][-1])

        # Test infeasible model:
        self.bd1.model.new_constr = pyomo.Constraint(expr=self.bd1.model.p_el_vars[0] == 0)
        self.bd1.model.p_el_vars[0].setub(15.0)
        with self.assertRaises(NonoptimalError):
            f.solve(full_update=True, debug=False)
        return

    def test_dual_decomposition(self):
        f = algorithms['dual-decomposition'](self.cd, eps_primal=0.001)
        f.solve()

        self.assertAlmostEqual(20, self.bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(20, self.bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[0], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[1], 4)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[1], 2)
        return

    def test_stand_alone_algorithm(self):
        f = algorithms['stand-alone'](self.cd)
        f.solve()

        self.assertAlmostEqual(20, self.bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(20, self.bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[0], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[1], 4)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[1], 2)
        return

    def test_local_algorithm(self):
        f = algorithms['local'](self.cd)
        f.solve()

        self.assertAlmostEqual(20, self.bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(20, self.bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[0], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[1], 4)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[1], 2)
        return

    def test_central_algorithm(self):
        f = algorithms['central'](self.cd)
        f.solve()

        self.assertAlmostEqual(20, self.bd1.p_el_schedule[0], 4)
        self.assertAlmostEqual(20, self.bd1.p_el_schedule[1], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[0], 4)
        self.assertAlmostEqual(40, self.bd2.p_el_schedule[1], 4)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[0], 2)
        self.assertAlmostEqual(60, self.cd.p_el_schedule[1], 2)
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import unittest
import datetime
import logging
import warnings
import pytest
import pyomo.environ as pyomo
from pyomo.opt import TerminationCondition
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
from shapely.geometry import Point

from pycity_scheduling import constants, solvers
from pycity_scheduling.classes import *
from pycity_scheduling.util.metric import *


class TestModule(unittest.TestCase):
    def test_filter_entities(self):
        e = get_env(4, 8)
        bd = Building(e)
        bes = BuildingEnergySystem(e)
        pv = Photovoltaic(e, 0)
        bes.addDevice(pv)
        bd.addEntity(bes)

        def do_test(gen):
            entities = list(gen)
            self.assertEqual(1, len(entities))
            self.assertIn(pv, entities)

        do_test(filter_entities(bd.get_entities(), 'PV'))
        do_test(filter_entities(bd, 'generation_devices'))
        do_test(filter_entities(bd, [Photovoltaic]))
        do_test(filter_entities(bd, ['PV']))
        do_test(filter_entities(bd, {'PV': Photovoltaic}))
        with self.assertRaises(ValueError):
            next(filter_entities(bd, 'PPV'))
        with self.assertRaises(ValueError):
            next(filter_entities(bd, [int]))
        with self.assertRaises(ValueError):
            next(filter_entities(bd, None))
        return


class TestBattery(unittest.TestCase):
    def setUp(self):
        e = get_env(3)
        self.bat = Battery(e, 10, 20, soc_init=0.875, eta=0.5)
        return

    def test_populate_model(self):
        model = pyomo.ConcreteModel()
        self.bat.populate_model(model)
        model.c1 = pyomo.Constraint(expr=self.bat.model.e_el_vars[2] == 10)
        model.c2 = pyomo.Constraint(expr=self.bat.model.e_el_vars[0] == 5)
        obj = 0
        for t in range(len(self.bat.model.p_el_demand_vars)):
            obj += self.bat.model.p_el_demand_vars[t] * self.bat.model.p_el_demand_vars[t]
        model.o = pyomo.Objective(expr=obj)
        result = solve_model(model)

        # TODO stats are currently not correct due to a pyomo bug
        # use result as a workaround
        #model.compute_statistics()
        #stats = model.statistics
        #self.assertEqual(12, stats.number_of_variables)
        self.assertEqual(13, result.Problem[0].number_of_variables)
        var_sum = pyomo.value(sum(self.bat.model.p_el_vars[t] for t in range(1, 3)))
        self.assertAlmostEqual(40, var_sum, places=5)
        var_sum = pyomo.value(sum(
            self.bat.model.p_el_supply_vars[t] + self.bat.model.p_el_demand_vars[t] for t in range(1, 3)
        ))
        self.assertAlmostEqual(40, var_sum, places=5)
        return

    def test_update_model(self):
        model = pyomo.ConcreteModel()
        self.bat.populate_model(model)
        demand_var = self.bat.model.p_el_vars
        self.bat.update_model()
        model.c1 = pyomo.Constraint(expr=self.bat.model.e_el_vars[0] == 10)
        obj = 0
        for t in range(len(demand_var)):
            obj += demand_var[t] * demand_var[t]
        model.o = pyomo.Objective(expr=obj)
        solve_model(model)

        self.assertAlmostEqual(10, pyomo.value(demand_var[0]), places=5)
        return

    def test_update_schedule(self):
        model = pyomo.ConcreteModel()
        self.bat.populate_model(model)
        self.bat.update_model()
        self.bat.model.p_el_demand_vars.setlb(3.0)
        self.bat.model.p_el_demand_vars.setub(3.0)
        self.bat.model.p_el_supply_vars.setlb(0.0)
        self.bat.model.p_el_supply_vars.setub(0.0)
        obj = 0
        for t in range(len(self.bat.model.p_el_demand_vars)):
            obj += self.bat.model.p_el_demand_vars[t] * self.bat.model.p_el_demand_vars[t]
        model.o = pyomo.Objective(expr=obj)
        solve_model(model)
        self.bat.update_schedule()
        assert_equal_array(self.bat.p_el_schedule, [3] * 3)
        assert_equal_array(self.bat.e_el_schedule, 0.875 * 10 + np.arange(1, 4)*3*0.25*0.5)
        return

    def test_calculate_co2(self):
        self.bat.p_el_schedule = np.array([10]*3)
        self.assertEqual(0, calculate_co2(self.bat))
        return

    def test_get_objective(self):
        from pyomo.core import expr as EXPR
        model = pyomo.ConcreteModel()
        self.bat.populate_model(model)
        obj = self.bat.get_objective(2)
        vs = list(EXPR.identify_variables(obj))
        for t in range(3):
            self.assertTrue(any(self.bat.model.p_el_vars[t] is v for v in vs))
            self.bat.model.p_el_vars[t] = t * 5
        self.assertEqual(3, len(vs))
        self.assertEqual(sum(2*(5*t)**2 for t in range(3)), pyomo.value(obj))
        return

    def test_e_ini(self):
        expected_schedule = list(range(4, 21, 2))
        e = get_env(3, 9, 2)
        model = pyomo.ConcreteModel()
        bat = Battery(e, 20, 10, soc_init=0.1, eta=0.8)
        bat.populate_model(model)
        model.o = pyomo.Objective(expr=-bat.model.e_el_vars[2])
        for t in range(4):
            bat.update_model()
            solve_model(model)
            bat.update_schedule()
            e.timer.mpc_update()
            assert_equal_array(bat.e_el_schedule, expected_schedule[:3+t*2] + [0] * 2 * (3-t))
            assert_equal_array(bat.p_el_schedule, [10] * (3 + t * 2) + [0] * 2 * (3 - t))
            assert_equal_array(bat.p_el_demand_schedule, [10] * (3 + t * 2) + [0] * 2 * (3 - t))
            assert_equal_array(bat.p_el_supply_schedule, [0] * 9)
        return

    def test_no_discharge(self):
        e = get_env(9, 9)
        model = pyomo.ConcreteModel()
        bat = Battery(e, 30, 10, p_el_max_discharge=0, soc_init=0.5, eta=1)
        bat.populate_model(model)
        bat.update_model()
        model.o = pyomo.Objective(expr=sum(bat.model.p_el_vars[t] for t in range(len(bat.model.p_el_vars))))
        solve_model(model)
        bat.update_schedule()
        assert_equal_array(bat.e_el_schedule, [15] * 9)
        assert_equal_array(bat.p_el_schedule, [0] * 9)
        assert_equal_array(bat.p_el_demand_schedule, [0] * 9)
        assert_equal_array(bat.p_el_supply_schedule, [0] * 9)
        return


class TestBoiler(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.bl = Boiler(e, 10, 0.4)
        return

    def test_calculate_co2(self):
        self.bl.p_th_heat_schedule = - np.array([10] * 8)
        self.bl.p_th_heat_ref_schedule = - np.array([4] * 8)
        co2_em = np.array([1111]*8)

        co2 = calculate_co2(self.bl, co2_emissions=co2_em)
        self.assertEqual(50.0*constants.CO2_EMISSIONS_GAS, co2)
        co2 = calculate_co2(self.bl, timestep=4, co2_emissions=co2_em)
        self.assertEqual(25.0*constants.CO2_EMISSIONS_GAS, co2)
        self.bl.load_schedule("ref")
        co2 = calculate_co2(self.bl, co2_emissions=co2_em)
        self.assertEqual(20.0*constants.CO2_EMISSIONS_GAS, co2)
        return

    def test_lower_activation(self):
        e = get_env(4, 8)
        bl = Boiler(e, 10, lower_activation_limit=0.5)
        model = pyomo.ConcreteModel()
        bl.populate_model(model, "integer")
        bl.update_model("integer")
        model.o = pyomo.Objective(expr=bl.model.p_th_heat_vars[0])
        results = solve_model(model)
        self.assertEqual(TerminationCondition.optimal, results.solver.termination_condition)
        bl.model.p_th_heat_vars[0].setub(-0.1)
        bl.model.p_th_heat_vars[0].setlb(-4.9)
        logger = logging.getLogger("pyomo.core")
        oldlevel = logger.level
        logger.setLevel(logging.ERROR)
        results = solve_model(model)
        logger.setLevel(oldlevel)
        self.assertEqual(TerminationCondition.infeasible, results.solver.termination_condition)
        return

    def test_objective(self):
        model = pyomo.ConcreteModel()
        self.bl.populate_model(model)
        self.bl.get_objective()
        return


class TestBuilding(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.bd = Building(e)
        return

    def test_get_objective(self):
        from pyomo.core import expr as EXPR
        model = pyomo.ConcreteModel()
        env = self.bd.environment
        env.prices.tou_prices[:4] = [1, 2, 3, 4]
        env.prices.co2_prices[:4] = [5, 4, 3, 2]
        bes = BuildingEnergySystem(env)
        self.bd.addEntity(bes)
        self.bd.populate_model(model)
        obj = self.bd.get_objective(2)
        vs = list(EXPR.identify_variables(obj))
        self.assertEqual(4, len(vs))
        for t in range(4):
            self.bd.model.p_el_vars[t].value = 10**t
        self.assertAlmostEqual(2*4321/10*4, pyomo.value(obj), places=5)

        model = pyomo.ConcreteModel()
        bd2 = Building(env, 'co2')
        bd2.addEntity(bes)
        bd2.populate_model(model)
        obj = bd2.get_objective(2)
        vs = list(EXPR.identify_variables(obj))
        self.assertEqual(4, len(vs))
        for t in range(4):
            bd2.model.p_el_vars[t].value = 10**t
        # numerical errors caused by /14 and co2_prices being np.float32
        self.assertAlmostEqual(0.0, pyomo.value(obj), places=3)

        model = pyomo.ConcreteModel()
        bd3 = Building(env, 'peak-shaving')
        bd3.addEntity(bes)
        bd3.populate_model(model)
        obj = bd3.get_objective(2)
        vs = list(EXPR.identify_variables(obj))
        self.assertEqual(4, len(vs))
        for t in range(4):
            bd3.model.p_el_vars[t].value = 10**t
        self.assertEqual(2*1010101, pyomo.value(obj))

        model = pyomo.ConcreteModel()
        bd4 = Building(env, None)
        bd4.addEntity(bes)
        bd4.populate_model(model)
        obj = bd4.get_objective(2)
        vs = list(EXPR.identify_variables(obj))
        self.assertEqual(0, len(vs))
        for t in range(4):
            bd4.model.p_el_vars[t].value = 10 ** t
        self.assertEqual(0, pyomo.value(obj))

        bd5 = Building(env, "invalid")
        self.assertRaisesRegex(ValueError, ".*Building.*", bd5.get_objective)
        return

    def test_calculate_co2(self):
        bes = BuildingEnergySystem(self.bd.environment)
        pv = Photovoltaic(self.bd.environment, 0)
        bes.addDevice(pv)
        self.bd.addEntity(bes)
        self.bd.p_el_schedule = np.array([-5] * 2 + [5] * 4 + [-5] * 2)
        self.bd.p_el_ref_schedule = np.array([-2] * 2 + [2] * 4 + [-2] * 2)
        pv.p_el_schedule = - np.array([10]*8)
        pv.p_el_ref_schedule = - np.array([4]*8)
        co2_em = np.array([100]*4 + [400]*4)

        co2 = calculate_co2(self.bd, co2_emissions=co2_em)
        self.assertEqual(20.0*constants.CO2_EMISSIONS_PV+1250.0, co2)
        co2 = calculate_co2(self.bd, timestep=4, co2_emissions=co2_em)
        self.assertEqual(10.0*constants.CO2_EMISSIONS_PV+250.0, co2)
        self.bd.load_schedule("ref")
        co2 = calculate_co2(self.bd, co2_emissions=co2_em)
        self.assertEqual(8.0*constants.CO2_EMISSIONS_PV+500.0, co2)
        return

    def test_robustness(self):
        model = pyomo.ConcreteModel()
        env = self.bd.environment
        bes = BuildingEnergySystem(env)
        self.bd.addEntity(bes)
        ths1 = ThermalHeatingStorage(env, 10)
        bes.addDevice(ths1)
        ths2 = ThermalHeatingStorage(env, 25)
        bes.addDevice(ths2)

        ap = Apartment(env)
        self.bd.addEntity(ap)
        loadcurve = np.array([15, 15, 10, 10])
        sh = SpaceHeating(env, loadcurve=loadcurve)
        ap.addEntity(sh)

        eh = ElectricHeater(env, 20)
        bes.addDevice(eh)

        self.bd.populate_model(model, robustness=(3, 0.5))
        self.bd.update_model(robustness=(3, 0.5))
        assert_equal_array(np.array([self.bd.model.lower_robustness_bounds[i].value for i in range(3)]),
                           np.cumsum(loadcurve[:3])*0.5/4)
        assert_equal_array(np.array([self.bd.model.upper_robustness_bounds[i].value for i in range(3)]),
                           35 - np.cumsum(loadcurve[:3]) * 0.5 / 4)

        self.assertEqual(17.5, self.bd.model.lower_robustness_bounds[3].value)
        self.assertEqual(17.5, self.bd.model.upper_robustness_bounds[3].value)
        return

    def testReset(self):
        env = self.bd.environment
        bes = BuildingEnergySystem(env)
        self.bd.addEntity(bes)
        schedules = list(self.bd.schedules.keys())
        model = pyomo.ConcreteModel()
        self.bd.populate_model(model)
        self.bd.update_model()
        model.o = pyomo.Objective(expr=sum(self.bd.model.p_el_vars[t] for t in range(len(self.bd.model.p_el_vars))))
        solve_model(model)
        self.assertEqual(schedules, list(self.bd.schedules.keys()))
        self.bd.update_schedule()
        self.assertEqual(schedules, list(self.bd.schedules.keys()))
        self.bd.schedules["ref"]["p_el"] = np.arange(8)
        self.bd.copy_schedule("new", "ref")
        schedules.append("new")
        self.bd.reset("ref")
        for k in schedules:
            if k == "new":
                e = np.arange(8)
            else:
                e = np.zeros(8)
            assert_equal_array(self.bd.schedules[k]["p_el"], e)
        self.bd.reset()
        for k in schedules:
            assert_equal_array(self.bd.schedules[k]["p_el"], np.zeros(8))
        self.assertEqual(schedules, list(self.bd.schedules.keys()))
        with self.assertRaises(KeyError):
            self.bd.load_schedule("nonexistent")
            self.bd.p_el_schedule
        with self.assertRaises(KeyError):
            self.bd.load_schedule(None)
            self.bd.p_el_schedule
        return


class TestChiller(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.ch = Chiller(e, 10, cop=np.full(8, 11))
        return

    def test_update_model(self):
        from pyomo.core import expr as EXPR
        m = pyomo.ConcreteModel()
        self.ch.populate_model(m)
        self.ch.update_model()

        c = self.ch.model.p_coupl_constr[0]
        f, l = EXPR.decompose_term(c.body)
        self.assertTrue(f)
        for coeff, value in l:
            if not np.isscalar(coeff):
                coeff = coeff.value
            if value is self.ch.model.p_el_vars[0]:
                self.assertEqual(11, coeff)
            if value is self.ch.model.p_th_cool_vars[0]:
                self.assertEqual(1, coeff)
            if value is None:
                self.assertEqual(0, coeff)
        return

    def test_lower_activation(self):
        e = get_env(4, 8)
        ch = Chiller(e, 10, cop=np.full(8, 11), lower_activation_limit=0.5)
        m = pyomo.ConcreteModel()
        ch.populate_model(m, "integer")
        ch.update_model("integer")
        obj = 0
        for t in range(len(ch.model.p_th_cool_vars)):
            obj += ch.model.p_th_cool_vars[t] * ch.model.p_th_cool_vars[t]
        obj += 2 * 3 * sum(ch.model.p_th_cool_vars[t] for t in range(len(ch.model.p_th_cool_vars)))
        m.o = pyomo.Objective(expr=obj)
        solve_model(m)
        ch.update_schedule()
        assert_equal_array(ch.p_th_cool_schedule[:4], [-5] * 4)
        return


class TestCurtailableLoad(unittest.TestCase):
    combinations = [(4, 1), (3, 1), (2, 1), (1, 1),
                    (1, 3), (1, 4), (2, 2), (2, 3),
                    (0, 1), (0, 2), (0, 3), (0, 4)]
    horizon = 5

    def setUp(self):
        self.e = get_env(5, 20)
        return

    def test_populate_model(self):
        model = pyomo.ConcreteModel()
        cl = CurtailableLoad(self.e, 2, 0.5)
        cl.populate_model(model)
        obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
        model.o = pyomo.Objective(expr=obj)
        solve_model(model)
        cl.update_schedule()
        self.assertAlmostEqual(5, pyomo.value(obj))
        self.assertTrue(5, sum(cl.p_el_schedule[:5]))
        return

    def test_populate_model_on_off(self):
        model = pyomo.ConcreteModel()
        cl = CurtailableLoad(self.e, 2, 0.5, 2, 2)
        cl.populate_model(model)
        obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
        model.o = pyomo.Objective(expr=obj)
        solve_model(model)
        cl.update_schedule()
        self.assertAlmostEqual(7, pyomo.value(obj), places=4)
        self.assertAlmostEqual(7, sum(cl.p_el_schedule[:5]), places=4)
        return

    def test_populate_model_integer(self):
        for low, full in self.combinations:
            min_states = sum(np.tile([False]*low + [True]*full, 5)[:5])
            for nom in [0.5, 1, 2]:
                with self.subTest(msg="max_low={} min_full={} nom={}".format(low, full, nom)):
                    model = pyomo.ConcreteModel()
                    cl = CurtailableLoad(self.e, nom, 0.75, low, full)
                    cl.populate_model(model, mode="integer")
                    obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                    model.o = pyomo.Objective(expr=obj)
                    results = solve_model(model)
                    cl.update_schedule()
                    schedule_states = np.isclose(cl.p_el_schedule[:5], [nom]*5)
                    assert_equal_array(cl.p_state_schedule[:5], schedule_states)
                    self.assertEqual(min_states, sum(schedule_states))
                    self.assertAlmostEqual(min_states*nom+(5-min_states)*nom*0.75, pyomo.value(obj))
        return

    def test_update_model(self):
        for width in [1, 2, 4, 5]:
            with self.subTest(msg="step width={}".format(width)):
                model = pyomo.ConcreteModel()
                cl = CurtailableLoad(self.e, 2, 0.5)
                cl.populate_model(model)
                obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                model.o = pyomo.Objective(expr=obj)
                solve_model(model)
                for t in range(0, 20-5+1, width):
                    self.e.timer.current_timestep = t
                    cl.update_model()
                    solve_model(model)
                    cl.update_schedule()
                    self.assertAlmostEqual(5, pyomo.value(obj))
                    self.assertAlmostEqual(5, sum(cl.p_el_schedule[t:t+5]))
        return

    def test_update_model_on_off(self):
        for low, full in self.combinations:
            for width in [1, 2, 4, 5]:
                with self.subTest(msg="max_low={} min_full={} step width={}".format(low, full, width)):
                    model = pyomo.ConcreteModel()
                    cl = CurtailableLoad(self.e, 2, 0.5, low, full)
                    cl.populate_model(model)
                    obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                    model.o = pyomo.Objective(expr=obj)
                    solve_model(model)
                    for t in range(0, 20-5+1, width):
                        self.e.timer.current_timestep = t
                        cl.update_model()
                        solve_model(model)
                        cl.update_schedule()

                    endtimestep = self.e.timer.current_timestep + cl.op_horizon
                    for t in range(0, endtimestep):
                        self.assertGreaterEqual(cl.p_el_schedule[t], 0.9999)
                        self.assertLessEqual(cl.p_el_schedule[t], 2.0001)
                    for t in range(0, endtimestep-(low+full)+1):
                        self.assertGreaterEqual(sum(cl.p_el_schedule[t:t+low+full]) + 1e-4, 1*low + 2*full)
        return

    def test_update_model_integer(self):
        for low, full in self.combinations:
            states = np.tile([False] * low + [True] * full, 20)[:20]
            for width in [1, 2, 4, 5]:
                with self.subTest(msg="max_low={} min_full={} step width={}".format(low, full, width)):
                    model = pyomo.ConcreteModel()
                    cl = CurtailableLoad(self.e, 2, 0.5, low, full)
                    cl.populate_model(model, mode="integer")
                    obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                    for t in range(0, 20-5+1, width):
                        self.e.timer.current_timestep = t
                        cl.update_model(mode="integer")
                        model.o = pyomo.Objective(expr=obj)
                        results = solve_model(model)
                        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
                        best_obj = pyomo.value(obj)
                        model.o_constr = pyomo.Constraint(expr=best_obj == obj)
                        model.del_component("o")
                        objective = 0
                        count = 0
                        for i in range(0, -cl.op_horizon, -1):
                            objective += i * cl.model.p_el_vars[count]
                            count += 1
                        model.o = pyomo.Objective(expr=objective)
                        results = solve_model(model)
                        model.del_component("o")
                        model.del_component("o_constr")
                        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
                        cl.update_schedule()
                        schedule_states_el = np.isclose(cl.p_el_schedule[t:t+5], [2] * 5)
                        schedule_states_b = np.isclose(cl.p_state_schedule[t:t+5], [1] * 5)
                        assert_equal_array(schedule_states_b, states[t:t + 5])
                        assert_equal_array(schedule_states_el, schedule_states_b)
                        assert_equal_array(
                            cl.p_el_schedule[t:t+5],
                            np.full(5, 2 * 0.5) + np.array(states[t:t+5]) * (2 * (1. - 0.5))
                        )
        return

    def test_integer_first(self):
        for low, full in self.combinations:
            if low > 0:
                with self.subTest(msg="max_low={} min_full={}".format(low, full)):
                    model = pyomo.ConcreteModel()

                    cl = CurtailableLoad(self.e, 2, 0.5, low, full)
                    cl.populate_model(model, mode="integer")
                    self.e.timer.current_timestep = 1
                    cl.p_state_schedule[0] = False
                    cl.p_el_schedule[0] = 1
                    cl.update_model("integer")

                    cl.model.p_state_vars[0].setub(1.0)
                    cl.model.p_state_vars[0].setlb(1.0)
                    cl.model.p_state_vars[1].setub(0.0)
                    cl.model.p_state_vars[1].setlb(0.0)

                    model.o = pyomo.Objective(expr=cl.model.p_state_vars[0])
                    logger = logging.getLogger("pyomo.core")
                    oldlevel = logger.level
                    logger.setLevel(logging.ERROR)
                    results = solve_model(model)
                    logger.setLevel(oldlevel)
                    if full > 1:
                        self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasible)
                    else:
                        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
        return

    def test_small_horizon(self):
        for width in [1, 2, 4]:
            for horizon in [1, 2, 4]:
                if horizon >= width:
                    with self.subTest(msg="width={} horizon={}".format(width, horizon)):
                        e = get_env(horizon, 20)
                        model = pyomo.ConcreteModel()
                        cl = CurtailableLoad(e, 2, 0.5)
                        cl.populate_model(model)
                        obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                        model.o = pyomo.Objective(expr=obj)
                        for t in range(0, 21 - horizon, width):
                            e.timer.current_timestep = t
                            cl.update_model()
                            solve_model(model)
                            self.assertEqual(1, pyomo.value(cl.model.p_el_vars[0]))
                            cl.update_schedule()
                        assert_equal_array(cl.p_el_schedule, [1] * 20)
        return

    def test_small_horizon_low_full(self):
        for horizon in [1, 2, 4]:
            e = get_env(horizon, 20)
            for width in [1, 2, 4]:
                if horizon >= width:
                    for low, full in self.combinations:
                        with self.subTest(msg="width={} horizon={} max_low={} min_full={}"
                                              .format(width, horizon, low, full)):

                            model = pyomo.ConcreteModel()
                            cl = CurtailableLoad(e, 2, 0.5, low, full)
                            cl.populate_model(model)
                            obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                            model.c = pyomo.Objective(expr=obj)
                            for t in range(0, 21 - horizon, width):
                                e.timer.current_timestep = t
                                cl.update_model()
                                solve_model(model)
                                cl.update_schedule()

                            for t in range(0, 20 - (low + full) + 1):
                                self.assertGreaterEqual(sum(cl.p_el_schedule[t:t + low + full]) + 1e-4,
                                                        1 * low + 2 * full,
                                                        np.array2string(cl.p_el_schedule))
        return

    def test_small_horizon_low_full_integer(self):
        for horizon in [1, 2, 4]:
            e = get_env(horizon, 20)
            for width in [1, 2, 4]:
                if horizon >= width:
                    for low, full in self.combinations:
                        with self.subTest(msg="width={} horizon={} max_low={} min_full={}".format(width, horizon,
                                                                                                  low, full)):
                            states = np.tile([1] * low + [2] * full, 20)[:20]
                            model = pyomo.ConcreteModel()
                            cl = CurtailableLoad(e, 2, 0.5, low, full)
                            cl.populate_model(model, mode="integer")
                            obj = sum(cl.model.p_el_vars[t] for t in range(len(cl.model.p_el_vars)))
                            for t in range(0, 21 - horizon, width):
                                e.timer.current_timestep = t
                                cl.update_model(mode="integer")
                                model.o = pyomo.Objective(expr=obj)
                                results = solve_model(model)
                                self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
                                best_obj = pyomo.value(obj)
                                model.o_constr = pyomo.Constraint(expr=best_obj == obj)
                                model.del_component("o")
                                objective = 0
                                count = 0
                                for i in range(-1, -cl.op_horizon-1, -1):
                                    objective += i * cl.model.p_el_vars[count]
                                    count += 1
                                model.o = pyomo.Objective(expr=objective)
                                results = solve_model(model)
                                model.del_component("o")
                                model.del_component("o_constr")
                                self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
                                cl.update_schedule()

                            assert_equal_array(cl.p_el_schedule, states)
        return


class TestCityDistrict(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.cd = CityDistrict(e)
        return

    def test_get_objective(self):
        m = pyomo.ConcreteModel()
        self.cd.populate_model(m)

        def zero_constr(model, t):
            return model.p_el_vars[t] == 0

        self.cd.model.extra_constr = pyomo.Constraint(self.cd.model.t, rule=zero_constr)
        m.o = pyomo.Objective(expr=self.cd.get_objective())
        solve_model(m)

        for t in range(4):
            self.cd.model.p_el_vars[t].value = t

        self.assertEqual(self.cd.objective, "price")
        self.cd.environment.prices.da_prices = np.array([1]*2 + [4]*6)
        self.assertAlmostEqual(8.4, pyomo.value(self.cd.get_objective()))
        self.cd.objective = 'peak-shaving'
        self.assertAlmostEqual(14, pyomo.value(self.cd.get_objective()))
        self.cd.objective = 'valley-filling'
        self.cd.valley_profile = np.array([-1]*8)
        self.assertAlmostEqual(2, pyomo.value(self.cd.get_objective()))
        self.cd.objective = None
        self.assertAlmostEqual(0, pyomo.value(self.cd.get_objective()))
        self.cd.objective = "invalid"
        self.assertRaisesRegex(ValueError, ".*CityDistrict.*", self.cd.get_objective)

        m = pyomo.ConcreteModel()
        self.cd.objective = "max-consumption"
        self.cd.populate_model(m)
        self.cd.model.p_el_vars[0].setub(-1)
        m.o = pyomo.Objective(expr=self.cd.get_objective())
        solve_model(m)
        self.assertAlmostEqual(1, pyomo.value(self.cd.get_objective()))
        return

    def test_calculate_costs(self):
        self.cd.p_el_schedule = np.array([10]*4 + [-20]*4)
        self.cd.p_el_ref_schedule = np.array([4]*4 + [-4]*4)
        prices = np.array([10]*4 + [20]*4)

        costs = calculate_costs(self.cd, prices=prices, feedin_factor=0.5)
        self.assertEqual(-100, costs)
        costs = calculate_costs(self.cd, timestep=4, prices=prices)
        self.assertEqual(100, costs)
        self.cd.load_schedule("ref")
        costs = calculate_costs(self.cd, prices=prices)
        self.assertEqual(-40, costs)
        return

    def test_calculate_co2(self):
        pv = Photovoltaic(self.cd.environment, 0)
        self.cd.addEntity(pv, Point(0, 0))
        self.cd.p_el_schedule = np.array([-5] * 2 + [5] * 4 + [-5] * 2)
        self.cd.p_el_ref_schedule = np.array([-2] * 2 + [2] * 4 + [-2] * 2)
        pv.p_el_schedule = - np.array([10] * 8)
        pv.p_el_ref_schedule = - np.array([4] * 8)
        co2_em = np.array([100] * 4 + [400] * 4)

        co2 = calculate_co2(self.cd, co2_emissions=co2_em)
        self.assertEqual(20.0*constants.CO2_EMISSIONS_PV+1250.0, co2)
        co2 = calculate_co2(self.cd, timestep=4, co2_emissions=co2_em)
        self.assertEqual(10.0*constants.CO2_EMISSIONS_PV+250.0, co2)
        self.cd.load_schedule("ref")
        co2 = calculate_co2(self.cd, co2_emissions=co2_em)
        self.assertEqual(8.0*constants.CO2_EMISSIONS_PV+500.0, co2)
        return

    def test_self_consumption(self):
        pv = Photovoltaic(self.cd.environment, 0)
        self.cd.addEntity(pv, Point(0, 0))
        self.cd.p_el_schedule = np.array([4]*2 + [-4]*2 + [-10]*2 + [-2]*2)
        self.cd.p_el_ref_schedule = np.array([2]*2 + [-6]*2 + [-9]*2 + [-1]*2)
        pv.p_el_schedule = - np.array([0]*2 + [8]*4 + [0]*2)
        pv.p_el_ref_schedule = - np.array([0]*8)

        self.assertEqual(0.25, self_consumption(self.cd))
        self.assertEqual(0.5, self_consumption(self.cd, timestep=4))
        self.cd.load_schedule("ref")
        self.assertEqual(1, self_consumption(self.cd))
        return

    def test_calculate_adj_costs(self):
        self.cd.p_el_schedule = np.array([4] * 2 + [-4] * 2 + [-10] * 2 + [-2] * 2)
        self.cd.p_el_ref_schedule = np.array([2] * 2 + [-6] * 2 + [-9] * 2 + [-1] * 2)
        prices = np.array([10] * 4 + [20] * 4)
        costs_adj = calculate_adj_costs(self.cd, "ref", prices=prices)
        self.assertEqual(2*5+2*5+1*10+1*10, costs_adj)
        costs_adj = calculate_adj_costs(self.cd, "ref", prices=prices, total_adjustments=False)
        self.assertEqual(20, costs_adj)
        self.cd.copy_schedule("ref")
        costs_adj = calculate_adj_costs(self.cd, "ref", prices=prices)
        self.assertEqual(0, costs_adj)
        return

    def test_autarky(self):
        pv = Photovoltaic(self.cd.environment, 0)
        self.cd.addEntity(pv, Point(0, 0))
        self.cd.p_el_schedule = np.array([4]*2 + [-4]*2 + [-10]*2 + [-2]*2)
        self.cd.p_el_ref_schedule = - np.array([0]*2 + [8]*4 + [0]*2)
        pv.p_el_schedule = - np.array([0]*2 + [8]*4 + [0]*2)
        pv.p_el_ref_schedule = - np.array([0]*2 + [8]*4 + [0]*2)

        self.assertEqual(0.5, autarky(self.cd))
        self.assertEqual(0, autarky(self.cd, timestep=2))
        self.cd.load_schedule("ref")
        self.assertEqual(1, autarky(self.cd))
        return


class TestCombinedHeatPower(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.chp = CombinedHeatPower(e, 10, 10, 0.8)
        return

    def test_calculate_co2(self):
        self.chp.p_th_heat_schedule = - np.array([10] * 8)
        self.chp.p_th_heat_ref_schedule = - np.array([4] * 8)
        co2_em = np.array([1111]*8)

        co2 = calculate_co2(self.chp, co2_emissions=co2_em)
        self.assertEqual(50.0*constants.CO2_EMISSIONS_GAS, co2)
        co2 = calculate_co2(self.chp, timestep=4, co2_emissions=co2_em)
        self.assertEqual(25.0*constants.CO2_EMISSIONS_GAS, co2)
        self.chp.load_schedule("ref")
        co2 = calculate_co2(self.chp, co2_emissions=co2_em)
        self.assertEqual(20.0*constants.CO2_EMISSIONS_GAS, co2)
        return

    def test_lower_activation(self):
        e = get_env(4, 8)
        chp = CombinedHeatPower(e, 10, 10, 0.8, 0.5)
        m = pyomo.ConcreteModel()
        chp.populate_model(m, "integer")
        chp.update_model("integer")
        obj = 0
        for t in range(len(chp.model.p_el_vars)):
            obj += chp.model.p_el_vars[t] * chp.model.p_el_vars[t]
        obj += 2*3 * sum(chp.model.p_el_vars[t] for t in range(len(chp.model.p_el_vars)))
        m.o = pyomo.Objective(expr=obj)
        solve_model(m)
        chp.update_schedule()
        assert_equal_array(chp.p_el_schedule[:4], [-5]*4)
        return


class TestDeferrableLoad(unittest.TestCase):
    def setUp(self):
        self.e = get_env(6, 9)
        self.lt = [0, 1, 1, 1, 0, 1, 1, 1, 0]
        return

    def test_update_model(self):
        with pytest.warns(UserWarning):
            dl = DeferrableLoad(self.e, 19, 10, load_time=self.lt)
        model = pyomo.ConcreteModel()
        dl.populate_model(model)
        obj = pyomo.sum_product(dl.model.p_el_vars, dl.model.p_el_vars)
        model.o = pyomo.Objective(expr=obj)
        dl.update_model()
        solve_model(model)

        self.assertAlmostEqual(10, pyomo.value(sum(dl.model.p_el_vars[t] for t in range(len(dl.model.p_el_vars))))
                               * dl.time_slot, places=5)

        dl.timer.mpc_update()
        dl.update_model()
        solve_model(model)

        for t, c in enumerate(self.lt[1:7]):
            if c == 1:
                self.assertGreaterEqual(19, dl.model.p_el_vars[t].value-1e-3)
            else:
                self.assertGreaterEqual(0, dl.model.p_el_vars[t].value-1e-3)
        dl.update_schedule()
        assert_equal_array(dl.p_el_schedule[:7], [0, 8, 8, 8, 0, 8, 8])
        return

    def test_large_runtime(self):
        self.e = get_env(9, 9)
        self.lt = [1, 1, 1, 1, 1, 1, 1, 1, 1]
        dl= DeferrableLoad(self.e, 10, 90/4, load_time=self.lt)
        m = pyomo.ConcreteModel()
        dl.populate_model(m)
        dl.update_model()
        obj = pyomo.sum_product(dl.model.p_el_vars)
        m.o = pyomo.Objective(expr=obj)
        results = solve_model(m)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)

        dl.update_schedule()
        assert_equal_array(dl.p_el_schedule, np.full_like(dl.p_el_schedule, 10))
        assert_equal_array(dl.p_start_schedule, [True, False, False, False, False, False, False, False, False])

        self.e = get_env(9, 9)
        self.lt = [1, 1, 1, 1, 1, 1, 1, 1, 1]
        dl = DeferrableLoad(self.e, 10, 90/4, load_time=self.lt)
        m = pyomo.ConcreteModel()
        dl.populate_model(m, mode="integer")
        dl.update_model()
        obj = pyomo.sum_product(dl.model.p_el_vars)
        m.o = pyomo.Objective(expr=obj)
        results = solve_model(m)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)

        dl.update_schedule()
        assert_equal_array(dl.p_el_schedule, np.full_like(dl.p_el_schedule, 10))
        assert_equal_array(dl.p_start_schedule, [True, False, False, False, False, False, False, False, False])
        return

    def test_small_runtime(self):
        self.e = get_env(9, 9)
        self.lt = [1, 1, 1, 1, 1, 1, 1, 1, 1]
        dl = DeferrableLoad(self.e, 10, 10/4, load_time=self.lt)
        m = pyomo.ConcreteModel()
        dl.populate_model(m)
        dl.update_model()
        obj = pyomo.sum_product(dl.model.p_el_vars, dl.model.p_el_vars)
        m.o = pyomo.Objective(expr=obj)
        results = solve_model(m)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)

        dl.update_schedule()
        assert_equal_array(dl.p_el_schedule, np.full_like(dl.p_el_schedule, 10/9))

        self.e = get_env(9, 9)
        self.lt = [1, 1, 1, 1, 1, 1, 1, 1, 1]
        dl = DeferrableLoad(self.e, 10, 10/4, load_time=self.lt)
        m = pyomo.ConcreteModel()
        dl.populate_model(m, mode="integer")
        dl.update_model()
        obj = pyomo.sum_product(dl.model.p_el_vars, dl.model.p_el_vars)
        m.o = pyomo.Objective(expr=obj)
        results = solve_model(m)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)

        dl.update_schedule()
        values, counts = np.unique(dl.p_el_schedule, return_counts=True)
        assert_equal_array(values, [0, 10])
        assert_equal_array(counts, [8, 1])
        return

    def test_zero_runtime(self):
        self.e = get_env(9, 9)
        self.lt = [1, 1, 1, 1, 1, 1, 1, 1, 1]
        with self.assertRaises(ValueError):
            dl = DeferrableLoad(self.e, 10, 1, load_time=self.lt)
        return

    def test_small_horizon(self):
        e = get_env(1, 9)
        lt = [1, 1, 1, 1, 1, 1, 1, 1, 0]
        with pytest.warns(UserWarning):
            dl = DeferrableLoad(e, 10, 10 / 4, lt)

        for mode in ["integer", "convex"]:
            with self.subTest("mode: {}".format(mode)):
                e.timer.reset()
                dl.reset()

                m = pyomo.ConcreteModel()

                dl.populate_model(m, mode=mode)

                obj = pyomo.sum_product(dl.model.p_el_vars)
                m.o = pyomo.Objective(expr=obj)

                schedule = np.zeros(9)
                logger = logging.getLogger("pyomo.core")
                oldlevel = logger.level
                logger.setLevel(logging.ERROR)
                for t in range(9):
                    if t < 7 and mode == "integer":
                        with pytest.warns(UserWarning):
                            dl.update_model(mode=mode)
                    else:
                        dl.update_model(mode=mode)

                    results = solve_model(m)

                    if t == 8:
                        if str(results.solver.termination_condition) == "infeasible":
                            self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasible)
                        else:
                            self.assertEqual(results.solver.termination_condition,
                                             TerminationCondition.infeasibleOrUnbounded)
                        with pytest.warns(UserWarning, match="stale variable"):
                            dl.update_schedule(verbose=True)
                    else:
                        schedule[t] = 10
                        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
                        dl.update_schedule()

                        assert_equal_array(dl.p_el_schedule, schedule)

                        dl.timer.mpc_update()

                logger.setLevel(oldlevel)
                return

    def test_infeasible_consumption(self):
        with pytest.warns(UserWarning):
            feasible = DeferrableLoad(self.e, 10, 10, load_time=self.lt)
        m = pyomo.ConcreteModel()
        feasible.populate_model(m)
        feasible.update_model()
        obj = sum(feasible.model.p_el_vars[t] for t in range(len(feasible.model.p_el_vars)))
        m.o = pyomo.Objective(expr=obj)
        results = solve_model(m)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)

        m = pyomo.ConcreteModel()
        with pytest.warns(UserWarning):
            infeasible = DeferrableLoad(self.e, 10, 10.6, load_time=self.lt)
        infeasible.populate_model(m)
        infeasible.update_model()
        obj = sum(infeasible.model.p_el_vars[t] for t in range(len(infeasible.model.p_el_vars)))
        m.o = pyomo.Objective(expr=obj)
        logger = logging.getLogger("pyomo.core")
        oldlevel = logger.level
        logger.setLevel(logging.ERROR)
        results = solve_model(m)
        logger.setLevel(oldlevel)
        if str(results.solver.termination_condition) == "infeasible":
            self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasible)
        else:
            self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasibleOrUnbounded)
        return

    def test_update_model_integer(self):
        with pytest.warns(UserWarning):
            dl = DeferrableLoad(self.e, 19, 9.5, load_time=self.lt)
        m = pyomo.ConcreteModel()
        dl.populate_model(m, mode="integer")

        coeff = [0] * 2 + [1] * 2 + [0] * 2
        obj = 0
        for t in range(len(dl.model.p_el_vars)):
            obj += coeff[t] * dl.model.p_el_vars[t] * dl.model.p_el_vars[t]
        m.o = pyomo.Objective(expr=obj)
        with pytest.warns(UserWarning):
            dl.update_model(mode="integer")
        results = solve_model(m)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
        dl.update_schedule()

        assert_equal_array(np.rint(dl.p_el_schedule[:6]), [0, 19, 19, 0, 0, 0])
        for t in range(3):
            dl.timer.mpc_update()
            if t == 0:
                with pytest.warns(UserWarning):
                    dl.update_model(mode="integer")
            else:
                dl.update_model(mode="integer")
            results = solve_model(m)
            self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
            dl.update_schedule()

        assert_equal_array(dl.p_el_schedule, [0, 19, 19, 0, 0, 0, 19, 19, 0])
        return

    def test_infeasible_integer(self):
        e = get_env(1, 9)
        model = pyomo.ConcreteModel()
        with pytest.raises(ValueError):
            with pytest.warns(UserWarning):
                dl = DeferrableLoad(e, 19, 9.5, load_time=self.lt)
            dl.populate_model(model, mode="integer")
            dl.update_model(mode="integer")

        model = pyomo.ConcreteModel()
        with pytest.warns(UserWarning):
            dl = DeferrableLoad(self.e, 19, 19, load_time=self.lt)
            dl.populate_model(model, mode="integer")
            dl.update_model(mode="integer")
        obj = sum(dl.model.p_el_vars[t] for t in range(len(dl.model.p_el_vars)))
        model.o = pyomo.Objective(expr=obj)
        logger = logging.getLogger("pyomo.core")
        oldlevel = logger.level
        logger.setLevel(logging.ERROR)
        results = solve_model(model)
        logger.setLevel(oldlevel)
        if str(results.solver.termination_condition) == "infeasible":
            self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasible)
        else:
            self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasibleOrUnbounded)

        model = pyomo.ConcreteModel()
        with pytest.warns(UserWarning):
            dl = DeferrableLoad(self.e, 19, 19*3/4, load_time=self.lt)
            dl.populate_model(model, mode="integer")
            dl.update_model(mode="integer")
        obj = sum(dl.model.p_el_vars[t] for t in range(len(dl.model.p_el_vars)))
        model.o = pyomo.Objective(expr=obj)
        results = solve_model(model)
        self.assertEqual(results.solver.termination_condition, TerminationCondition.optimal)
        dl.update_schedule()
        assert_equal_array(dl.p_el_schedule[:6], [0, 19, 19, 19, 0, 0])
        return

    def test_objective(self):
        with pytest.warns(UserWarning):
            dl = DeferrableLoad(self.e, 19, 19, load_time=self.lt)
        model = pyomo.ConcreteModel()
        dl.populate_model(model)
        dl.get_objective()
        return

    def test_update_integer(self):
        e = get_env(9, 9)
        model = pyomo.ConcreteModel()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", UserWarning)
            dl = DeferrableLoad(e, 19, 19, load_time=[1] * 9)
            dl.populate_model(model, "integer")
            dl.update_model("integer")
            self.assertEqual(0, len(w))
        return


class TestFixedLoad(unittest.TestCase):
    def test_populate_model(self):
        e = get_env(2, 4)
        load = np.arange(1, 5)
        model = pyomo.ConcreteModel()
        self.fl = FixedLoad(e, method=0, demand=load)
        self.fl.populate_model(model)
        self.fl.update_model()
        model.o = pyomo.Objective(expr=sum(self.fl.model.p_el_vars[t] for t in range(len(self.fl.model.p_el_vars))))
        solve_model(model)
        for t in range(2):
            self.assertEqual(self.fl.model.p_el_vars[t].value, load[t])
        return

    def test_unit_conversion(self):
        ti = Timer(step_size=1800,
                   op_horizon=48,
                   mpc_horizon=24*365,
                   mpc_step_width=1)
        we = Weather(ti)
        pr = Prices(ti)
        e = Environment(ti, we, pr)
        fl = FixedLoad(e, method=1, annual_demand=25)
        # loadcurve in Wh
        self.assertEqual(48 * 365, len(fl.loadcurve))
        self.assertAlmostEqual(25*1000, sum(fl.loadcurve)*0.5, places=5)
        # p_el_schedule in kWh
        self.assertEqual(24*365, len(fl.p_el_schedule))
        self.assertAlmostEqual(25/2, sum(fl.p_el_schedule)*0.5, delta=25/2/10)
        return


class TestElectricalEntity(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8, 4)
        self.ee = ElectricalEntity(e)
        self.ee.environment = e
        return

    def test_update_schedule(self):
        m = pyomo.ConcreteModel()
        self.ee.populate_model(m)
        for t in range(4):
            self.ee.model.p_el_vars[t].value = t
        a = np.arange(4)

        self.ee.update_schedule()
        assert_equal_array(self.ee.p_el_schedule[:4], a)
        self.ee.timer.mpc_update()
        self.ee.update_schedule()
        assert_equal_array(self.ee.p_el_schedule[4:], a)
        return

    def test_calculate_costs(self):
        self.ee.p_el_schedule = np.array([10]*4 + [-20]*4)
        self.ee.p_el_ref_schedule = np.array([4]*4 + [-4]*4)
        prices = np.array([10]*4 + [20]*4)

        costs = calculate_costs(self.ee, prices=prices, feedin_factor=0.5)
        self.assertEqual(-100, costs)
        costs = calculate_costs(self.ee, timestep=4, prices=prices)
        self.assertEqual(100, costs)
        self.ee.load_schedule("ref")
        costs = calculate_costs(self.ee, prices=prices)
        self.assertEqual(40, costs)
        return

    def test_calculate_adj_costs(self):
        self.ee.p_el_schedule = np.array([10] * 4 + [-20] * 4)
        self.ee.p_el_ref_schedule = np.array([4] * 4 + [-4] * 4)
        prices = np.array([10] * 4 + [20] * 4)
        costs_adj = calculate_adj_costs(self.ee, "ref", prices=prices)
        self.assertEqual(6*10 + 16*20, costs_adj)
        costs_adj = calculate_adj_costs(self.ee, "ref", prices=prices, total_adjustments=False)
        self.assertEqual(16 * 20, costs_adj)
        self.ee.copy_schedule("ref")
        costs_adj = calculate_adj_costs(self.ee, "ref", prices=prices)
        self.assertEqual(0, costs_adj)
        return

    def test_calculate_adj_power(self):
        self.ee.p_el_schedule = np.array([10] * 4 + [-20] * 4)
        self.ee.p_el_ref_schedule = np.array([4] * 4 + [-4] * 4)
        adj_power = calculate_adj_power(self.ee, "ref")
        assert_equal_array(adj_power, [6] * 4 + [16] * 4)
        adj_power = calculate_adj_power(self.ee, "ref", total_adjustments=False)
        assert_equal_array(adj_power, [0] * 4 + [16] * 4)
        adj_power = calculate_adj_power(self.ee, "default")
        assert_equal_array(adj_power, [0] * 8)
        self.ee.load_schedule("ref")
        adj_power = calculate_adj_power(self.ee, "ref")
        assert_equal_array(adj_power, [0] * 8)
        self.ee.copy_schedule("default")
        adj_power = calculate_adj_power(self.ee, "default")
        assert_equal_array(adj_power, [0] * 8)
        return

    def test_calculate_adj_energy(self):
        self.ee.p_el_schedule = np.array([10] * 4 + [-20] * 4)
        self.ee.p_el_ref_schedule = np.array([4] * 4 + [-4] * 4)
        adj_energy = calculate_adj_energy(self.ee, "ref")
        self.assertEqual(6 + 16, adj_energy)
        adj_energy = calculate_adj_energy(self.ee, "ref", total_adjustments=False)
        self.assertEqual(16, adj_energy)
        adj_energy = calculate_adj_energy(self.ee, "default")
        self.assertEqual(0, adj_energy)
        self.ee.copy_schedule(src="ref")
        adj_energy = calculate_adj_energy(self.ee, "ref")
        self.assertEqual(0, adj_energy)
        adj_energy = calculate_adj_energy(self.ee, "ref", total_adjustments=False)
        self.assertEqual(0, adj_energy)
        self.ee.load_schedule("ref")
        adj_energy = calculate_adj_energy(self.ee, "ref")
        self.assertEqual(0, adj_energy)
        adj_energy = calculate_adj_energy(self.ee, "default")
        self.assertEqual(0, adj_energy)
        return

    def test_metric_delta_g(self):
        self.ee.p_el_schedule = np.array([10] * 4 + [-20] * 4)
        self.ee.p_el_ref_schedule = np.array([4] * 4 + [-4] * 4)
        g = metric_delta_g(self.ee, "ref")
        self.assertEqual(1-30/8, g)
        g = metric_delta_g(self.ee, "default")
        self.assertEqual(0, g)
        return

    def test_peak_to_average_ratio(self):
        self.ee.p_el_schedule = np.array([10] * 4 + [-20] * 4)
        self.ee.p_el_ref_schedule = np.array([4] * 4 + [-4] * 4)
        ratio = peak_to_average_ratio(self.ee)
        self.assertEqual(20/5, ratio)
        self.ee.load_schedule("ref")
        with pytest.warns(RuntimeWarning):
            ratio = peak_to_average_ratio(self.ee)
        self.assertEqual(np.inf, ratio)
        return

    def test_peak_reduction_ratio(self):
        self.ee.p_el_schedule = np.array([10] * 4 + [-20] * 4)
        self.ee.p_el_ref_schedule = np.array([4] * 4 + [-4] * 4)
        ratio = peak_reduction_ratio(self.ee, "ref")
        self.assertEqual((20-4)/4, ratio)
        self.ee.p_el_ref_schedule = np.array([4] * 8)
        ratio = peak_reduction_ratio(self.ee, "ref")
        self.assertEqual((20-4)/4, ratio)
        ratio = peak_reduction_ratio(self.ee, "default")
        self.assertEqual(0, ratio)
        self.ee.load_schedule("ref")
        ratio = peak_reduction_ratio(self.ee, "ref")
        self.assertEqual(0, ratio)
        return

    def test_self_consumption(self):
        # properly tested in CityDistrict
        self.ee.p_el_schedule = np.array([10]*4 + [-20]*4)
        self.assertEqual(0, self_consumption(self.ee))
        return

    def test_autarky(self):
        # properly tested in CityDistrict
        self.ee.p_el_schedule = np.array([10]*4 + [-20]*4)
        self.assertEqual(0, autarky(self.ee))
        return

    def test_objective(self):
        model = pyomo.ConcreteModel()
        self.ee.populate_model(model)
        self.ee.get_objective()
        return

    def test_new_objective(self):
        model = pyomo.ConcreteModel()
        self.ee.populate_model(model)
        for t in range(4):
            self.ee.model.p_el_vars[t].setlb(t)
            self.ee.model.p_el_vars[t].setub(t)
        self.ee.set_objective("peak-shaving")
        obj = self.ee.get_objective()
        model.o = pyomo.Objective(expr=obj)
        solve_model(model)
        obj = self.ee.get_objective()
        self.assertEqual(sum(t**2 for t in range(4)), pyomo.value(obj))
        self.ee.set_objective("max-consumption")
        with self.assertRaises(ValueError):
            obj = self.ee.get_objective()

        model = pyomo.ConcreteModel()
        self.ee.populate_model(model)
        for t in range(4):
            self.ee.model.p_el_vars[t].setlb(t)
            self.ee.model.p_el_vars[t].setub(t)
        obj = self.ee.get_objective()
        model.o = pyomo.Objective(expr=obj)
        solve_model(model)

        self.assertAlmostEqual(9.0, pyomo.value(obj), 4)
        return


class TestElectricHeater(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.eh = ElectricHeater(e, 10, 10, 0.8)
        return

    def test_lower_activation(self):
        e = get_env(4, 8)
        eh = ElectricHeater(e, 10, lower_activation_limit=0.5)
        m = pyomo.ConcreteModel()
        eh.populate_model(m, "integer")
        eh.update_model("integer")
        obj = 0
        for t in range(len(eh.model.p_el_vars)):
            obj += eh.model.p_el_vars[t] * eh.model.p_el_vars[t]
        obj += -2 * 3 * sum(eh.model.p_el_vars[t] for t in range(len(eh.model.p_el_vars)))
        m.o = pyomo.Objective(expr=obj)
        solve_model(m)
        eh.update_schedule()
        assert_equal_array(eh.p_el_schedule[:4], [5] * 4)
        return

    def test_update_schedule(self):
        e = get_env(2, 2)
        eh = ElectricHeater(e, 10, lower_activation_limit=0.5)
        m = pyomo.ConcreteModel()
        eh.populate_model(m)
        eh.update_model()
        obj = eh.model.p_el_vars[0] - eh.model.p_el_vars[1]
        eh.model.p_el_vars[0].setlb(5.0)
        eh.model.p_el_vars[1].setub(0.05)
        m.o = pyomo.Objective(expr=obj)
        solve_model(m)
        eh.update_schedule()
        assert_equal_array(eh.p_el_schedule, [5, 0.05])
        assert_equal_array(eh.p_th_heat_schedule, [-5, -0.05])
        assert_equal_array(eh.p_th_heat_state_schedule, [True, False])
        return


class TestElectricVehicle(unittest.TestCase):
    def setUp(self):
        e = get_env(6, 9)
        self.ct = [1, 1, 1, 0, 0, 0, 1, 1, 1]
        self.ev = ElectricVehicle(e, 10, 20, p_el_max_discharge=20, soc_init=0.5,
                                  charging_time=self.ct, simulate_driving=True, minimum_soc_end=1.0, eta=1.0)
        return

    def test_populate_model(self):
        model = pyomo.ConcreteModel()
        self.ev.populate_model(model)
        model.c1 = pyomo.Constraint(expr=self.ev.model.e_el_vars[2] == 10)
        model.c2 = pyomo.Constraint(expr=self.ev.model.e_el_vars[0] == 5)
        obj = 0
        for t in range(len(self.ev.model.p_el_demand_vars)):
            obj += self.ev.model.p_el_demand_vars[t] * self.ev.model.p_el_demand_vars[t]
        model.o = pyomo.Objective(expr=obj)
        result = solve_model(model)

        self.assertEqual(31, result.Problem[0].number_of_variables)
        var_sum = pyomo.value(sum(self.ev.model.p_el_vars[t] for t in range(1, 6)))
        self.assertAlmostEqual(20, var_sum, places=2)
        var_sum = pyomo.value(sum(
            self.ev.model.p_el_supply_vars[t] + self.ev.model.p_el_demand_vars[t] for t in range(1, 6)))
        self.assertAlmostEqual(20, var_sum, places=2)
        return

    def test_update_model(self):
        model = pyomo.ConcreteModel()
        self.ev.populate_model(model)
        self.ev.update_model()
        model.o = pyomo.Objective(expr=self.ev.get_objective())
        solve_model(model)

        self.assertAlmostEqual(10, self.ev.model.e_el_vars[2].value, places=5)
        self.assertAlmostEqual(10, self.ev.model.e_el_vars[3].value, places=5)

        self.ev.timer.mpc_update()
        self.ev.update_model()
        solve_model(model)

        for t, c in enumerate(self.ct[1:7]):
            if c:
                self.assertGreaterEqual(20, self.ev.model.p_el_demand_vars[t].value-1e-3)
                self.assertGreaterEqual(20, self.ev.model.p_el_supply_vars[t].value-1e-3)
                self.assertGreaterEqual(0, self.ev.model.p_el_drive_vars[t].value-1e-3)
            else:
                self.assertGreaterEqual(0, self.ev.model.p_el_demand_vars[t].value-1e-3)
                self.assertGreaterEqual(0, self.ev.model.p_el_supply_vars[t].value-1e-3)
                self.assertGreaterEqual(constants.INFINITY, self.ev.model.p_el_drive_vars[t].value+1e-3)
        self.assertAlmostEqual(10, self.ev.model.e_el_vars[1].value, places=5)
        self.assertAlmostEqual(10, self.ev.model.e_el_vars[2].value, places=5)
        self.assertLessEqual(3, self.ev.model.e_el_vars[5].value)
        return

    def test_get_objective(self):
        model = pyomo.ConcreteModel()
        self.ev.populate_model(model)
        self.ev.update_model()

        obj = self.ev.get_objective(11)
        for i in range(6):
            ref = (i + 1) / 21 * 6 * 11
            coeff = obj.args[i].args[0].args[0]
            self.assertAlmostEqual(ref, coeff, places=5)
        return

    def test_no_charge_time(self):
        e = get_env(6, 9)
        ev = ElectricVehicle(e, 37.0, 11.0)
        assert_equal_array(ev.charging_time, [1]*9)
        e = get_env(28, 96*24-12)
        ev = ElectricVehicle(e, 37.0, 11.0)
        assert_equal_array(ev.charging_time, np.tile([1] * 24 + [0] * 48 + [1] * 24, 24)[:-12])
        return

    def test_no_discharge(self):
        model = pyomo.ConcreteModel()
        e = get_env(6, 9)
        ev = ElectricVehicle(e, 10.0, 40.0, charging_time=self.ct, simulate_driving=True,
                             minimum_soc_end=1.0, eta=1.0)
        ev.populate_model(model)
        ev.update_model()
        model.o = pyomo.Objective(expr=ev.model.p_el_vars[0] + ev.model.p_el_vars[1])
        solve_model(model)
        ev.update_schedule()
        assert_equal_array(ev.p_el_schedule[:4], [0, 0, 5*4, 0])
        assert_equal_array(ev.p_el_demand_schedule[:4], [0, 0, 5 * 4, 0])
        assert_equal_array(ev.p_el_supply_schedule[:4], [0, 0, 0, 0])
        assert_equal_array(ev.e_el_schedule[:4], [5, 5, 10, 10])

        model = pyomo.ConcreteModel()
        e = get_env(6, 9)
        ev = ElectricVehicle(e, 10.0, 40.0, p_el_max_discharge=8, charging_time=self.ct,
                             simulate_driving=True, minimum_soc_end=1.0, eta=1.0)
        ev.populate_model(model)
        ev.update_model()
        model.o = pyomo.Objective(expr=ev.model.p_el_vars[0] + ev.model.p_el_vars[1])
        solve_model(model)
        ev.update_schedule()
        assert_equal_array(ev.p_el_schedule[:4], [0, 0, 20, 0])
        assert_equal_array(ev.p_el_demand_schedule[:4], [0, 8, 20, 0])
        assert_equal_array(ev.p_el_supply_schedule[:4], [0, 8, 0, 0])
        assert_equal_array(ev.e_el_schedule[:4], [5, 5, 10, 10])
        return

    def test_bad_charging_times(self):
        e = get_env(3, 12)
        self.ev = ElectricVehicle(e, 10, 8, soc_init=0.5,
                                  charging_time=[1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0],
                                  simulate_driving=True, minimum_soc_end=1.0, eta=1.0)
        with pytest.warns(UserWarning):
            self.ev = ElectricVehicle(e, 10, 8, soc_init=0.5,
                                      charging_time=[1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
                                      simulate_driving=True, minimum_soc_end=1.0, eta=1.0)
        with pytest.warns(UserWarning):
            self.ev = ElectricVehicle(e, 10, 8, soc_init=0.5,
                                      charging_time=[1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0],
                                      simulate_driving=True, minimum_soc_end=1.0, eta=1.0)
        with pytest.warns(UserWarning):
            self.ev = ElectricVehicle(e, 10, 8, soc_init=0.5,
                                      charging_time=[1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1],
                                      simulate_driving=True, minimum_soc_end=1.0, eta=1.0)
        return


class TestHeatPump(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.hp = HeatPump(e, 10, cop=np.full(8, 11))
        return

    def test_update_model(self):
        from pyomo.core import expr as EXPR
        m = pyomo.ConcreteModel()
        self.hp.populate_model(m)
        self.hp.update_model()

        c = self.hp.model.p_coupl_constr[0]
        f, l = EXPR.decompose_term(c.body)
        self.assertTrue(f)
        for coeff, value in l:
            if not np.isscalar(coeff):
                coeff = coeff.value
            if value is self.hp.model.p_el_vars[0]:
                self.assertEqual(11, coeff)
            if value is self.hp.model.p_th_heat_vars[0]:
                self.assertEqual(1, coeff)
            if value is None:
                self.assertEqual(0, coeff)
        return

    def test_lower_activation(self):
        e = get_env(4, 8)
        hp = HeatPump(e, 10, lower_activation_limit=0.5)
        m = pyomo.ConcreteModel()
        hp.populate_model(m, "integer")
        hp.update_model("integer")
        obj = 0
        for t in range(len(hp.model.p_th_heat_vars)):
            obj += hp.model.p_th_heat_vars[t] * hp.model.p_th_heat_vars[t]
        obj += 2 * 3 * sum(hp.model.p_th_heat_vars[t] for t in range(len(hp.model.p_th_heat_vars)))
        m.o = pyomo.Objective(expr=obj)
        solve_model(m)
        hp.update_schedule()
        assert_equal_array(hp.p_th_heat_schedule[:4], [-5] * 4)
        return


class TestPhotovoltaic(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.pv = Photovoltaic(e, 0, 30, 0.0, 0.3)
        return

    def test_calculate_co2(self):
        self.pv.p_el_schedule = - np.array([10]*8)
        self.pv.p_el_ref_schedule = - np.array([4]*8)
        co2_em = np.array([1111]*8)

        co2 = calculate_co2(self.pv, co2_emissions=co2_em)
        self.assertEqual(20.0*constants.CO2_EMISSIONS_PV, co2)
        co2 = calculate_co2(self.pv, timestep=4, co2_emissions=co2_em)
        self.assertEqual(10.0*constants.CO2_EMISSIONS_PV, co2)
        self.pv.load_schedule("ref")
        co2 = calculate_co2(self.pv, co2_emissions=co2_em)
        self.assertEqual(8*constants.CO2_EMISSIONS_PV, co2)
        return

    def test_objective(self):
        model = pyomo.ConcreteModel()
        self.pv.populate_model(model)
        self.pv.get_objective()
        return


class TestPrices(unittest.TestCase):
    def test_cache(self):
        Prices.co2_price_cache = None
        Prices.da_price_cache = None
        Prices.tou_price_cache = None
        Prices.tou_price_cache_year = None
        ti = Timer(op_horizon=4, mpc_horizon=8, step_size=3600,
                   initial_date=(2015, 1, 1), initial_time=(1, 0, 0))
        pr = Prices(ti)

        self.assertEqual(35040, len(pr.da_price_cache))
        self.assertEqual(35040, len(pr.tou_price_cache))
        self.assertEqual(35040, len(pr.co2_price_cache))
        self.assertTrue(np.allclose(pr.tou_prices, [23.2621]*6 + [42.2947]*2))

        Prices.da_price_cache[4] = 20
        ti = Timer(op_horizon=4, mpc_horizon=8, step_size=900,
                   initial_date=(2015, 1, 1), initial_time=(1, 0, 0))
        pr = Prices(ti)

        self.assertAlmostEqual(20, pr.da_prices[0], places=4)
        return

    def test_unavailable_year(self):
        ti = Timer(op_horizon=4, mpc_horizon=8, step_size=3600,
                   initial_date=(9999, 1, 1), initial_time=(1, 0, 0))
        with pytest.warns(UserWarning, match=r'9999'):
            Prices(ti)
        return


class TestThermalCoolingStorage(unittest.TestCase):
    def setUp(self):
        e = get_env(3)
        self.tcs = ThermalCoolingStorage(e, 40, 0.5)
        return

    def test_update_schedule(self):
        m = pyomo.ConcreteModel()
        self.tcs.populate_model(m)
        self.tcs.update_model()
        for t in range(3):
            self.tcs.model.p_th_cool_vars[t].setub(t)
            self.tcs.model.p_th_cool_vars[t].setlb(t)
        m.o = pyomo.Objective(expr=sum(self.tcs.model.p_th_cool_vars[t]
                                       for t in range(len(self.tcs.model.p_th_cool_vars))))
        solve_model(m)
        a = np.arange(3)

        self.tcs.update_schedule()
        assert_equal_array(self.tcs.p_th_cool_schedule, a)
        assert_equal_array(self.tcs.e_th_cool_schedule, [20, 20.25, 20.75])
        return


class TestThermalHeatingStorage(unittest.TestCase):
    def setUp(self):
        e = get_env(3)
        self.ths = ThermalHeatingStorage(e, 40, 0.5)
        return

    def test_update_schedule(self):
        m = pyomo.ConcreteModel()
        self.ths.populate_model(m)
        self.ths.update_model()
        for t in range(3):
            self.ths.model.p_th_heat_vars[t].setub(t)
            self.ths.model.p_th_heat_vars[t].setlb(t)
        m.o = pyomo.Objective(expr=sum(self.ths.model.p_th_heat_vars[t]
                                       for t in range(len(self.ths.model.p_th_heat_vars))))
        solve_model(m)
        a = np.arange(3)

        self.ths.update_schedule()
        assert_equal_array(self.ths.p_th_heat_schedule, a)
        assert_equal_array(self.ths.e_th_heat_schedule, [20, 20.25, 20.75])
        return


class TestThermalEntityCooling(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8, 4)
        self.tc = ThermalEntityCooling(e)
        self.tc.environment = e
        return

    def test_update_schedule(self):
        m = pyomo.ConcreteModel()
        self.tc.populate_model(m)
        self.tc.update_model()
        for t in range(4):
            self.tc.model.p_th_cool_vars[t].setub(t)
            self.tc.model.p_th_cool_vars[t].setlb(t)
        m.o = pyomo.Objective(expr=sum(self.tc.model.p_th_cool_vars[t]
                                       for t in range(len(self.tc.model.p_th_cool_vars))))
        solve_model(m)
        a = np.arange(4)

        self.tc.update_schedule()
        assert_equal_array(self.tc.p_th_cool_schedule[:4], a)
        self.tc.timer.mpc_update()
        self.tc.update_schedule()
        assert_equal_array(self.tc.p_th_cool_schedule[4:], a)
        return


class TestThermalEntityHeating(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8, 4)
        self.th = ThermalEntityHeating(e)
        self.th.environment = e
        return

    def test_update_schedule(self):
        m = pyomo.ConcreteModel()
        self.th.populate_model(m)
        self.th.update_model()
        for t in range(4):
            self.th.model.p_th_heat_vars[t].setub(t)
            self.th.model.p_th_heat_vars[t].setlb(t)
        m.o = pyomo.Objective(expr=sum(self.th.model.p_th_heat_vars[t]
                                       for t in range(len(self.th.model.p_th_heat_vars))))
        solve_model(m)
        a = np.arange(4)

        self.th.update_schedule()
        assert_equal_array(self.th.p_th_heat_schedule[:4], a)
        self.th.timer.mpc_update()
        self.th.update_schedule()
        assert_equal_array(self.th.p_th_heat_schedule[4:], a)
        return


class TestSpaceCooling(unittest.TestCase):
    def setUp(self):
        e = get_env(2, 4)
        self.load = np.arange(1, 5)
        self.sc = SpaceCooling(e, method=0, loadcurve=self.load)
        return

    def test_model(self):
        m = pyomo.ConcreteModel()
        self.sc.populate_model(m)
        self.sc.update_model()

        m.o = pyomo.Objective(expr=self.sc.model.p_th_cool_vars[0]+self.sc.model.p_th_cool_vars[1])
        r = solve_model(m)
        assert_equal_array(self.sc.p_th_cool_schedule, self.load)
        self.assertAlmostEqual(self.load[0], self.sc.model.p_th_cool_vars[0].value)
        self.assertAlmostEqual(self.load[1], self.sc.model.p_th_cool_vars[1].value)
        return


class TestSpaceHeating(unittest.TestCase):
    def setUp(self):
        e = get_env(2, 4)
        self.load = np.arange(1, 5)
        self.sh = SpaceHeating(e, method=0, loadcurve=self.load)
        return

    def test_model(self):
        m = pyomo.ConcreteModel()
        self.sh.populate_model(m)
        self.sh.update_model()

        m.o = pyomo.Objective(expr=self.sh.model.p_th_heat_vars[0]+self.sh.model.p_th_heat_vars[1])
        r = solve_model(m)
        assert_equal_array(self.sh.p_th_heat_schedule, self.load)
        self.assertAlmostEqual(self.load[0], self.sh.model.p_th_heat_vars[0].value)
        self.assertAlmostEqual(self.load[1], self.sh.model.p_th_heat_vars[1].value)
        return


class TestTimer(unittest.TestCase):
    def setUp(self):
        self.timer = Timer(mpc_horizon=192, mpc_step_width=4,
                           initial_date=(2015, 1, 15), initial_time=(12, 0, 0))
        self.timer._dt = datetime.datetime(2015, 1, 15, 13)
        return

    def test_time_in_year(self):
        self.assertEqual(1396, self.timer.time_in_year())
        self.assertEqual(1392, self.timer.time_in_year(from_init=True))
        return

    def test_time_in_week(self):
        self.assertEqual(340, self.timer.time_in_week())
        self.assertEqual(336, self.timer.time_in_week(from_init=True))
        return

    def test_time_in_day(self):
        self.assertEqual(52, self.timer.time_in_day())
        self.assertEqual(48, self.timer.time_in_day(from_init=True))
        return

    def test_more_than_one_year(self):
        for s, h, horizon in [(s, h, horizon) for s in [300, 900, 1800, 3600]
                              for h in range(int(86400*365/s)-1, int(86400*365/s)+3)
                              for horizon in ["op_horizon", "mpc_horizon"]]:
            year_horizon = int(86400 * 365 / s)
            kwargs = {"step_size": s, "initial_date": (2015, 1, 15), "initial_time": (12, 0, 0)}
            kwargs[horizon] = h
            with self.subTest(msg="step_size={} horizon={} horizon_name={}".format(s, h, horizon)):
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always", UserWarning)
                    t = Timer(**kwargs)
                    we = Weather(t)
                    if year_horizon < h:
                        self.assertEqual(len(w), 1,
                                         msg="No warning was thrown even though mpc_horizon / op_horizon {} "
                                             "is larger than one year, which would be a horizon of {}"
                                         .format(h, year_horizon))
                        self.assertIn(horizon, str(w[0].message))
                        self.assertEqual(len(we.p_ambient), year_horizon)
                        self.assertEqual(len(we.phi_ambient), year_horizon)
                        self.assertEqual(len(we.q_diffuse), year_horizon)
                        self.assertEqual(len(we.q_direct), year_horizon)
                        self.assertEqual(len(we.rad_earth), year_horizon)
                        self.assertEqual(len(we.rad_sky), year_horizon)
                        self.assertEqual(len(we.v_wind), year_horizon)
                        self.assertEqual(len(we.t_ambient), year_horizon)
                        self.assertEqual(len(we.current_p_ambient), year_horizon)
                        self.assertEqual(len(we.current_phi_ambient), year_horizon)
                        self.assertEqual(len(we.current_q_diffuse), year_horizon)
                        self.assertEqual(len(we.current_q_direct), year_horizon)
                        self.assertEqual(len(we.current_rad_earth), year_horizon)
                        self.assertEqual(len(we.current_rad_sky), year_horizon)
                        self.assertEqual(len(we.current_v_wind), year_horizon)
                        self.assertEqual(len(we.current_t_ambient), year_horizon)
                    else:
                        self.assertEqual(len(w), 0)
        return


class TestWindEnergyConverter(unittest.TestCase):
    def setUp(self):
        e = get_env(4, 8)
        self.wec = WindEnergyConverter(e, np.array([0, 10]), np.array([0, 10]))
        return

    def test_calculate_co2(self):
        self.wec.p_el_schedule = - np.array([10] * 8)
        self.wec.p_el_ref_schedule = - np.array([4] * 8)
        co2_em = np.array([1111]*8)

        co2 = calculate_co2(self.wec, co2_emissions=co2_em)
        self.assertEqual(20.0*constants.CO2_EMISSIONS_WIND, co2)
        co2 = calculate_co2(self.wec, timestep=4, co2_emissions=co2_em)
        self.assertEqual(10.0*constants.CO2_EMISSIONS_WIND, co2)
        self.wec.load_schedule("ref")
        co2 = calculate_co2(self.wec, co2_emissions=co2_em)
        self.assertEqual(8.0*constants.CO2_EMISSIONS_WIND, co2)
        return

    def test_objective(self):
        model = pyomo.ConcreteModel()
        self.wec.populate_model(model)
        self.wec.get_objective()
        return


def get_env(op_horizon, mpc_horizon=None, mpc_step_width=1):
    ti = Timer(op_horizon=op_horizon,
               mpc_horizon=mpc_horizon,
               mpc_step_width=mpc_step_width)
    we = Weather(ti)
    pr = Prices(ti)
    return Environment(ti, we, pr)


def assert_equal_array(a: np.ndarray, expected):
    if not np.allclose(a, expected, rtol=1e-3, atol=1e-6):
        expected = np.array(expected)
        msg = "Array {} does not equal expected array {}".format(np.array2string(a), np.array2string(expected))
        raise AssertionError(msg)


def solve_model(model):
    # hack to suppress pyomo no constraint warning
    if not hasattr(model, "simple_var"):
        model.simple_var = pyomo.Var(domain=pyomo.Reals, initialize=0)
        model.simple_constr = pyomo.Constraint(expr=model.simple_var == 1)
    opt = pyomo.SolverFactory(solvers.DEFAULT_SOLVER, **solvers.DEFAULT_SOLVER_OPTIONS.get("__call__", {}))
    if isinstance(opt, PersistentSolver):
        opt.set_instance(model, **solvers.DEFAULT_SOLVER_OPTIONS.get("set_instance", {}))
    return opt.solve(model, **solvers.DEFAULT_SOLVER_OPTIONS.get("solve", {}))



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec

import pycity_scheduling.util.factory as factory
from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a simple power scheduling example to demonstrate the integration and impact of building cooling and
# heating loads using the central optimization algorithm.


def main(do_plot=False):
    print("\n\n------ Example 22: Scheduling Heating and Cooling Loads ------\n\n")

    # Scheduling will be performed for a full year:
    env = factory.generate_standard_environment(step_size=3600, op_horizon=24*365, mpc_horizon=None,
                                                mpc_step_width=None, initial_date=(2018, 1, 1), initial_time=(0, 0, 0))

    # Use a standardized thermal load profile for space heating and then 'convert' it into a cooling load
    # (inverted load assumption according to pycity_base):
    sh_slp = SpaceHeating(environment=env, method=1, living_area=150, specific_demand=50)

    sc_slp = SpaceCooling(environment=env,
                          method=0,
                          loadcurve=np.ones(env.timer.timesteps_total)*max(sh_slp.get_power()) / 1000.0 - sh_slp.
                          get_power() / 1000.0)

    # Plot the thermal loads:
    plot_time = list(range(env.timer.timesteps_used_horizon))

    gs = gridspec.GridSpec(2, 1)
    ax0 = plt.subplot(gs[0])
    ax0.plot(plot_time, sh_slp.p_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.ylabel("Space Heating [kW]")

    ax1 = plt.subplot(gs[1], sharex=ax0)
    ax1.plot(plot_time, sc_slp.p_th_cool_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Space Cooling [kW]")
    plt.title("Thermal Load Profiles")
    plt.grid()
    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()

    # Now perform the scheduling of two buildings, where the first building is equipped with the heating load and
    # a thermal heating system and the second building is equipped with the cooling load and a thermal cooling system.

    # City district with district operator objective "price":
    cd = CityDistrict(environment=env, objective='max-consumption')

    # Building no. one with building objective price:
    bd1 = Building(environment=env, objective='price')
    cd.addEntity(entity=bd1, position=[0, 0])
    bes = BuildingEnergySystem(environment=env)
    bd1.addEntity(bes)
    ths = ThermalHeatingStorage(environment=env, e_th_max=16.0, soc_init=0.5, loss_factor=0.05)
    bes.addDevice(ths)
    hp = HeatPump(environment=env, p_th_nom=8.0)
    bes.addDevice(hp)
    ap = Apartment(environment=env)
    bd1.addEntity(ap)
    ap.addEntity(sh_slp)

    # Building no. two with building objective co2:
    bd2 = Building(environment=env, objective='co2')
    cd.addEntity(entity=bd2, position=[0, 1])
    bes = BuildingEnergySystem(environment=env)
    bd2.addEntity(bes)
    tcs = ThermalCoolingStorage(environment=env, e_th_max=10.0, soc_init=0.5, loss_factor=0.1)
    bes.addDevice(tcs)
    ch = Chiller(environment=env, p_th_nom=5.0, cop=3.5)
    bes.addDevice(ch)
    ap = Apartment(environment=env)
    bd2.addEntity(ap)
    ap.addEntity(sc_slp)

    # Perform the scheduling:
    opt = CentralOptimization(city_district=cd)
    results = opt.solve()
    cd.copy_schedule("central")

    # Plot the (thermal) schedules of interest:
    plot_time = list(range(env.timer.timesteps_used_horizon))

    gs = gridspec.GridSpec(4, 1)
    ax0 = plt.subplot(gs[0])
    ax0.plot(plot_time, -hp.p_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.ylim([-15, 15])
    plt.grid()
    plt.title("Thermal Schedules")
    plt.ylabel("Heat Pump [kW]")

    ax1 = plt.subplot(gs[1], sharex=ax0)
    ax1.plot(plot_time, ths.e_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Heating Storage [kWh]")

    ax2 = plt.subplot(gs[2], sharex=ax0)
    ax2.plot(plot_time, -ch.p_th_cool_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Chiller [kW]")

    ax3 = plt.subplot(gs[3], sharex=ax0)
    ax3.plot(plot_time, tcs.e_th_cool_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Cooling Storage [kWh]")

    plt.xlabel("Time", fontsize=12)

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()
    return


# Conclusions:
# In analogy to the previous example (compare example_16_scheduling_pv+battery_system.py), also thermal energy storage
# units can be used as a buffer inside a sector-coupled energy system to provide flexibility to the local electrical
# system. In this example, for instance, the thermal heating and cooling storage units are used to (partly) decouple the
# heating respectively cooling system's thermal generation from its electrical consumption. Thus, it becomes evident
# that one can take advantage from the thermal flexibility to achieve the given local and system level scheduling
# objectives, too.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)
